import json

import pytest
import responses

from datarobot import Project, BlenderModel, ModelJob


@responses.activate
@pytest.mark.usefixtures('client')
def test_blend_return_job_id():
    job_id = 1
    location = 'https://host_name.com/projects/p-id/modelJobs/{}/'.format(job_id)
    model1 = '57e27abc8e43553e572f8df7'
    model2 = '57e27abc8e43553e572f8df8'
    responses.add(responses.POST, 'https://host_name.com/projects/p-id/blenderModels/',
                  status=202,
                  body='',
                  adding_headers={'Location': location})
    model_job_url = 'https://host_name.com/projects/p-id/modelJobs/{}/'.format(job_id)
    model_job_data = {
        'status': 'inprogress',
        'processes': [],
        'projectId': 'p-id',
        'samplePct': 28.349,
        'modelType': 'Naive Bayes combiner classifier',
        'featurelistId': '56d8620bccf94e26f37af0a3',
        'modelCategory': 'model',
        'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
        'id': job_id
    }
    responses.add(responses.GET, model_job_url,
                  status=200,
                  body=json.dumps(model_job_data),
                  content_type='application/json')
    result = Project('p-id').blend([model1, model2], 'AVG')
    assert isinstance(result, ModelJob)
    assert result.id == job_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_blenders_return_valid_objects(blenders_list_response_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/blenderModels/',
                  body=blenders_list_response_json,
                  status=200,
                  content_type='application/json')
    blenders = Project('p-id').get_blenders()
    assert all(isinstance(blender, BlenderModel) for blender in blenders)



# coding: utf-8
from datetime import datetime
import json

import dateutil
import pytest
import responses

from datarobot.enums import JOB_TYPE
from datarobot.models.model import PrimeModel
from datarobot.models.prime_file import PrimeFile
from datarobot.models.ruleset import Ruleset
from ..utils import assert_equal_py2, assert_equal_py3


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_prime_model(project_url, project_id, prime_model_id,
                         prime_model_server_data, ruleset_with_model_server_data):
    responses.add(responses.GET, '{}primeModels/{}/'.format(project_url, prime_model_id),
                  body=json.dumps(prime_model_server_data))
    prime_model = PrimeModel.get(project_id, prime_model_id)

    assert prime_model.id == prime_model_server_data['id']
    assert prime_model.model_type == prime_model_server_data['modelType']
    assert prime_model.parent_model_id == ruleset_with_model_server_data['parentModelId']
    assert prime_model.ruleset.score == ruleset_with_model_server_data['score']


@responses.activate
@pytest.mark.usefixtures('client', 'approximate_job_creation_response')
def test_approximate(one_model, job_url, approximate_job_running_server_data):
    responses.add(responses.GET, job_url, body=json.dumps(approximate_job_running_server_data),
                  content_type='application/json')
    approx_job = one_model.request_approximation()
    assert approx_job.job_type == JOB_TYPE.PRIME_RULESETS


@pytest.fixture
def datetime_prime_data(datetime_model_data, prime_data):
    modified_data = dict(prime_data)
    modified_data.update(datetime_model_data)
    modified_data.pop('sample_pct')
    return modified_data


def test_datetime_prime_model(datetime_prime_data):
    mod = PrimeModel.from_data(datetime_prime_data)
    dt = mod.training_start_date
    assert dt.tzinfo == dateutil.tz.tzutc()
    assert isinstance(mod.training_end_date, datetime)
    assert mod.training_row_count is None
    assert mod.training_duration is None


@responses.activate
@pytest.mark.usefixtures('client',
                         'approximate_job_creation_response', 'approximate_completed_response')
def test_retrieve_approximate_result(one_model, prime_model_id):
    approx_job = one_model.request_approximation()
    rulesets = approx_job.get_result_when_complete()
    assert rulesets
    assert {prime_model_id, None} == {ruleset.model_id for ruleset in rulesets}
    for ruleset in rulesets:
        assert isinstance(ruleset, Ruleset)
        assert ruleset.parent_model_id == one_model.id


@responses.activate
@pytest.mark.usefixtures('client')
def test_check_primeable(project_url, one_model):
    response_data = {'canMakePrime': True, 'message': 'OK', 'messageId': 0}
    url = '{}models/{}/primeInfo/'.format(project_url, one_model.id)
    responses.add(responses.GET, url, body=json.dumps(response_data),
                  content_type='application/json')
    eligibility = one_model.get_prime_eligibility()
    assert eligibility == {'can_make_prime': True, 'message': 'OK'}


@responses.activate
@pytest.mark.usefixtures('client')
def test_check_primeable_future_proof(project_url, one_model):
    response_data = {'canMakePrime': True, 'message': 'OK', 'messageId': 0, 'future': 'new'}
    url = '{}models/{}/primeInfo/'.format(project_url, one_model.id)
    responses.add(responses.GET, url, body=json.dumps(response_data),
                  content_type='application/json')
    eligibility = one_model.get_prime_eligibility()
    assert eligibility == {'can_make_prime': True, 'message': 'OK'}


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_rulesets(project_url, model_id, prime_model_id, one_model,
                       ruleset_with_model_server_data, ruleset_without_model_server_data):
    response_data = [ruleset_with_model_server_data, ruleset_without_model_server_data]
    responses.add(responses.GET, '{}models/{}/primeRulesets/'.format(project_url, model_id),
                  body=json.dumps(response_data))
    rulesets = one_model.get_rulesets()

    assert len(rulesets) == 2
    ruleset_models = {ruleset.model_id for ruleset in rulesets}
    assert {prime_model_id, None} == ruleset_models


def test_future_proof(prime_model_server_data):
    future_data = dict(prime_model_server_data, future='new')
    PrimeModel.from_server_data(future_data)


@responses.activate
@pytest.mark.usefixtures('client', 'prime_validation_job_creation_response')
def test_validate(prime_model, job_url, prime_validation_job_running_server_data):
    responses.add(responses.GET, job_url, body=json.dumps(prime_validation_job_running_server_data),
                  content_type='application/json')
    validate_job = prime_model.request_download_validation('Python')
    assert validate_job.job_type == JOB_TYPE.PRIME_VALIDATION


@responses.activate
@pytest.mark.usefixtures('client', 'prime_validation_job_creation_response',
                         'prime_validation_job_completed_response')
def test_retrieve_validate_result(prime_model):
    validate_job = prime_model.request_download_validation('Python')
    prime_file = validate_job.get_result_when_complete()
    assert isinstance(prime_file, PrimeFile)
    assert prime_file.parent_model_id == prime_model.parent_model_id
    assert prime_file.language == 'Python'


class TestRepr(object):
    """Tests of method PrimeModel.__repr__"""

    @pytest.fixture
    def kwargs(self):
        """Required parameters of PrimeModel that don't affect the tests"""
        return {'ruleset_id': 1,
                'rule_count': 1,
                'score': 1.0,
                'parent_model_id': '58668f2fbf36cd0183a0617a',
                'project_id': '58668f44bf36cd0183a0617b'}

    def test_model_type_str(self, kwargs):
        # when both model_type and id are set
        frozen_model = PrimeModel(id='58668cb5bf36cd77a5d8056c', model_type='model_type', **kwargs)
        # then repr contains only model_type
        assert repr(frozen_model) == "PrimeModel('model_type')"

    def test_model_type_unicode(self, kwargs):
        # when both model_type and id are set
        # and model_type contains non-ascii characters
        frozen_model = PrimeModel(id='58668cb5bf36cd77a5d8056c', model_type=u'тайп', **kwargs)
        # then repr is an ascii string containing only model_type
        assert_equal_py2(repr(frozen_model), "PrimeModel(u'\\u0442\\u0430\\u0439\\u043f')")
        assert_equal_py3(repr(frozen_model), "PrimeModel('тайп')")

    def test_id(self, kwargs):
        # when id is set and model_type is missing
        frozen_model = PrimeModel(id='58668cb5bf36cd77a5d8056c', **kwargs)
        # then repr contains id
        assert repr(frozen_model) == "PrimeModel('58668cb5bf36cd77a5d8056c')"



import trafaret as t

from datarobot.models.api_object import APIObject


class PrimeFile(APIObject):
    """ Represents an executable file available for download of the code for a DataRobot Prime model

    Attributes
    ----------
    id : str
        the id of the PrimeFile
    project_id : str
        the id of the project this PrimeFile belongs to
    parent_model_id : str
        the model being approximated by this PrimeFile
    model_id : str
        the prime model this file represents
    ruleset_id : int
        the ruleset being used in this PrimeFile
    language : str
        the language of the code in this file - see enums.LANGUAGE for possibilities
    is_valid : bool
        whether the code passed basic validation
    """
    _converter = t.Dict({t.Key('id'): t.String(),
                         t.Key('project_id'): t.String(),
                         t.Key('parent_model_id'): t.String(),
                         t.Key('model_id'): t.String(),
                         t.Key('ruleset_id'): t.Int(),
                         t.Key('language'): t.String(),
                         t.Key('is_valid'): t.Bool()}).allow_extra('*')

    def __init__(self, id=None, project_id=None, parent_model_id=None, model_id=None,
                 ruleset_id=None, language=None, is_valid=None):
        self.id = id
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id
        self.ruleset_id = ruleset_id
        self.language = language
        self.is_valid = is_valid

    @classmethod
    def get(cls, project_id, file_id):
        url = 'projects/{}/primeFiles/{}/'.format(project_id, file_id)
        return cls.from_location(url)

    def download(self, filepath):
        """ Download the code and save it to a file

        Parameters
        ----------
        filepath: string
            the location to save the file to
        """
        url = 'projects/{}/primeFiles/{}/download/'.format(self.project_id, self.id)
        response = self._client.get(url)
        with open(filepath, mode='wb') as out_file:
            out_file.write(response.content)



import unittest
import contextlib
import warnings
import re
import sys

from datarobot.client import set_client
from datarobot.rest import RESTClientObject


class SDKTestcase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        set_client(None)
        c = RESTClientObject(auth='t-token', endpoint='https://host_name.com')
        set_client(c)

    @classmethod
    def tearDownClass(cls):
        set_client(None)


def assertJsonEq(first, second, msg='Ouups'):
    assert sorted(first) == sorted(second), '%r != %r\n%s' % (first,
                                                              second,
                                                              msg)


@contextlib.contextmanager
def warns(*categories):
    """Assert warnings are raised in a block, analogous to pytest.raises."""
    with warnings.catch_warnings(record=True) as messages:
        yield messages
    assert tuple(message.category for message in messages) == categories


def assert_raised_regex(exc_info, regexp):
    """
    Until pytest 2.10 when this gets implemented as a method of exc_info,
    we will provide this assertion.

    This is the exact implementation that has already been merged to pytest

    Parameters
    ----------
    exc_info : pytest.ExceptionInfo
    regexp

    Returns
    -------

    """
    if not re.search(regexp, str(exc_info.value)):
        assert 0, "Pattern '{0!s}' not found in '{1!s}'".format(
            regexp, exc_info.value
        )
    return True


def assert_equal_py2(this, that):
    """Assert that arguments are equal on Python 2.

    This function does nothing on Python 3.
    """
    if sys.version_info < (3, 0):
        assert this == that


def assert_equal_py3(this, that, message=''):
    """Assert that arguments are equal on Python 3.

    This function does nothing on Python 2.
    """
    if sys.version_info >= (3, 0):
        assert this == that



import json

import pytest
import responses

from datarobot.client import get_client

import mock


@pytest.yield_fixture
def client_version_mock():
    with mock.patch('datarobot.client._get_client_version') as mv:
        yield mv


class TestVersionCheck(object):

    @pytest.yield_fixture
    def server_version_mock(self):
        with mock.patch('datarobot.client._get_server_version') as mv:
            yield mv

    @pytest.yield_fixture(autouse=True)
    def no_global_client(self, default_config):
        with mock.patch('datarobot.client._global_client', None) as gc:
            yield gc

    def test_major_incompat(self, server_version_mock, client_version_mock):
        server_version_mock.return_value = '2.3.0'
        client_version_mock.return_value = '3.0.0'
        with pytest.raises(RuntimeError):
            get_client()

    @pytest.mark.usefixtures('known_warning')
    def test_minor_incompat(self, server_version_mock, client_version_mock):
        server_version_mock.return_value = '2.3.0'
        client_version_mock.return_value = '2.4.0'
        get_client()

    def test_server_can_be_ahead_of_client(self, server_version_mock, client_version_mock):
        server_version_mock.return_value = '2.4.0'
        client_version_mock.return_value = '2.3.0'
        get_client()


class TestServerVersion(object):

    @pytest.fixture
    def version_endpoint(self, default_config_endpoint):
        return default_config_endpoint + '/version/'

    @responses.activate
    @pytest.mark.usefixtures('default_config', 'known_warning')
    def test_version_does_not_exist(self, version_endpoint):
        """This situation happens if a user is using a client that is new
        enough to check for version on a server that doesn't have an endpoint
        to supply that version
        """
        responses.add(
            responses.GET,
            version_endpoint,
            status=404,
            body=''
        )

        get_client()

    @responses.activate
    @pytest.mark.usefixtures('default_config')
    def test_version_does_exist(self, version_endpoint, client_version_mock):
        client_version_mock.return_value = '2.4.0'

        responses.add(
            responses.GET,
            version_endpoint,
            status=200,
            body=json.dumps({'major': '2', 'minor': '4', 'versionString': '2.4.0'})
        )

        get_client()



import pytest
import responses

from datarobot import errors


@responses.activate
@pytest.mark.usefixtures('client',
                         'feature_impact_job_creation_response',
                         'feature_impact_running_response')
def test_get_feature_impact_job_result_not_finished(one_model):
    feature_impact_job = one_model.request_feature_impact()
    with pytest.raises(errors.JobNotFinished):
        feature_impact_job.get_result()


@responses.activate
@pytest.mark.usefixtures('client',
                         'feature_impact_job_creation_response',
                         'feature_impact_completed_response',
                         'feature_impact_response')
def test_get_feature_impact_job_result_finished(feature_impact_server_data, one_model):
    feature_impact_job = one_model.request_feature_impact()
    feature_impact = feature_impact_job.get_result()
    assert feature_impact == feature_impact_server_data['featureImpacts']


@responses.activate
@pytest.mark.usefixtures('client',
                         'feature_impact_job_creation_response',
                         'feature_impact_running_response')
def test_wait_for_feature_impact_never_finished(one_model, mock_async_time):
    mock_async_time.time.side_effect = (0, 5)
    feature_impact_job = one_model.request_feature_impact()
    with pytest.raises(errors.AsyncTimeoutError):
        feature_impact_job.get_result_when_complete(max_wait=1)


@responses.activate
@pytest.mark.usefixtures('client',
                         'feature_impact_job_creation_response',
                         'feature_impact_completed_response',
                         'feature_impact_response')
def test_wait_for_feature_impact_finished(one_model, feature_impact_server_data):
    # Would like to include a mock response for not done yet, but that didn't work.
    # Looks like `responses` might do that when `assert_all_requests_are_fired` is true.
    feature_impact_job = one_model.request_feature_impact()
    feature_impact = feature_impact_job.get_result_when_complete(max_wait=.5)
    assert feature_impact == feature_impact_server_data['featureImpacts']


@responses.activate
@pytest.mark.usefixtures('client',
                         'feature_impact_response')
def test_get_feature_impact_assumed_complete(one_model, feature_impact_server_data):
    feature_impact = one_model.get_feature_impact()
    assert feature_impact == feature_impact_server_data['featureImpacts']



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import get_id_from_response


class Ruleset(APIObject):
    """ Represents an approximation of a model with DataRobot Prime

    Attributes
    ----------
    id : str
        the id of the ruleset
    rule_count : int
        the number of rules used to approximate the model
    score : float
        the validation score of the approximation
    project_id : str
        the project the approximation belongs to
    parent_model_id : str
        the model being approximated
    model_id : str or None
        the model using this ruleset (if it exists).  Will be None if no such model has been
        trained.

    """
    _converter = t.Dict({t.Key('project_id'): t.String(),
                         t.Key('parent_model_id'): t.String(),
                         t.Key('model_id', optional=True): t.String(),
                         t.Key('ruleset_id'): t.Int(),
                         t.Key('rule_count'): t.Int(),
                         t.Key('score'): t.Float()}).allow_extra('*')

    def __init__(self, project_id=None, parent_model_id=None, model_id=None,
                 ruleset_id=None, rule_count=None, score=None):
        self.id = ruleset_id
        self.rule_count = rule_count
        self.score = score
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id

    def __repr__(self):
        return 'Ruleset(rule_count={}, score={})'.format(self.rule_count, self.score)

    def request_model(self):
        """ Request training for a model using this ruleset

        Training a model using a ruleset is a necessary prerequisite for being able to download
        the code for a ruleset.

        Returns
        -------
        job: Job
            the job fitting the new Prime model
        """
        from . import Job
        if self.model_id:
            raise ValueError('Model already exists for ruleset')
        url = 'projects/{}/primeModels/'.format(self.project_id)
        data = {'parent_model_id': self.parent_model_id, 'ruleset_id': self.id}
        response = self._client.post(url, data=data)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)



import json
import os
import tempfile
import warnings

import mock
import pytest
import responses

from datarobot import (Project, Model, PrimeFile, PrimeModel, Ruleset, enums, BlenderModel,
                       FrozenModel, ReasonCodesInitialization, ReasonCodes)
from datarobot.models.reason_codes import ReasonCodesPage
from datarobot.rest import RESTClientObject
from datarobot.utils import from_api
from datarobot.client import get_client, set_client


# This filter causes the tests to fail if any uncaught warnings leak out
warnings.simplefilter('error')


@pytest.yield_fixture
def temporary_file():
    new_file = tempfile.NamedTemporaryFile(delete=False)
    new_file.close()
    yield new_file.name
    os.remove(new_file.name)


@pytest.yield_fixture
def mock_async_time():
    patch = mock.patch('datarobot.async.time')
    yield patch.start()
    patch.stop()


@pytest.fixture
def unicode_string():
    return u'\u3053\u3093\u306b\u3061\u306f'


@pytest.fixture
def unittest_endpoint():
    return 'https://host_name.com'


@pytest.yield_fixture(scope='function')
def known_warning():
    """
    A context that will not let any warnings out. This will allow us to
    test for known deprecations and the like while still making sure the rest of the tested
    code does not emit any warnings

    This fixture asserts that a warning was raised while it was used, so users of this
    fixture don't need to do so themselves
    """
    filters = warnings.filters[:]
    warnings.resetwarnings()
    with warnings.catch_warnings(record=True) as w:
        yield w
    assert w, 'Expected a warning but did not find one'
    warnings.filters = filters


@pytest.yield_fixture(scope='function')
def client():
    """A mocked client

    The DataRobot package is built around acquiring the GlobalClient, which this sets
    to point to `https://host_name.com`.

    Most often you only need the effect, not the client. In this case you can use
    the pytest.mark.usefixtures decorator to make sure this patch is takes place
    for your test
    """
    c = RESTClientObject(auth='t-token', endpoint='https://host_name.com')
    set_client(c)
    yield get_client()
    set_client(None)


@pytest.fixture
def project_id():
    """
    A project id that matches the objects in the fixtures

    Returns
    -------
    project_id : str
        The id of the project used across the fixtures
    """
    return '556cdfbb100d2b0e88585195'


@pytest.fixture
def project_collection_url():
    return 'https://host_name.com/projects/'


@pytest.fixture
def project_url(project_id):
    return 'https://host_name.com/projects/{}/'.format(project_id)


@pytest.fixture
def project_aim_url(project_url):
    return '{}{}/'.format(project_url, 'aim')


@pytest.fixture
def project_without_target_json():
    """The JSON of one project

    This data in this project has been uploaded and analyzed, but the target
    has not been set
    """
    return """
    {
    "id": "556cdfbb100d2b0e88585195",
    "projectName": "A Project Name",
    "fileName": "test_data.csv",
    "stage": "aim",
    "autopilotMode": null,
    "created": "2016-07-26T02:29:58.546312Z",
    "target": null,
    "metric": null,
    "partition": {
      "datetimeCol": null,
      "cvMethod": null,
      "validationPct": null,
      "reps": null,
      "cvHoldoutLevel": null,
      "holdoutLevel": null,
      "userPartitionCol": null,
      "validationType": null,
      "trainingLevel": null,
      "partitionKeyCols": null,
      "holdoutPct": null,
      "validationLevel": null
    },
    "recommender": {
      "recommenderItemId": null,
      "isRecommender": null,
      "recommenderUserId": null
    },
    "advancedOptions": {
      "blueprintThreshold": null,
      "responseCap": null,
      "seed": null,
      "weights": null,
      "smartDownsampled": null,
      "majorityDownsamplingRate": null,
      "offset": null,
      "exposure": null
    },
    "positiveClass": null,
    "maxTrainPct": null,
    "holdoutUnlocked": false,
    "targetType": null
  }
"""


@pytest.fixture
def project_without_target_data(project_without_target_json):
    data = json.loads(project_without_target_json)
    return from_api(data)


@pytest.fixture
def project_with_target_json():
    return """{
        "id": "556cdfbb100d2b0e88585195",
        "projectName": "A Project Name",
        "fileName": "data.csv",
        "stage": "modeling",
        "autopilotMode": 0,
        "created": "2016-07-26T02:29:58.546312Z",
        "target": "target_name",
        "metric": "LogLoss",
        "partition": {
            "datetimeCol": null,
            "cvMethod": "stratified",
            "validationPct": null,
            "reps": 5,
            "cvHoldoutLevel": null,
            "holdoutLevel": null,
            "userPartitionCol": null,
            "validationType": "CV",
            "trainingLevel": null,
            "partitionKeyCols": null,
            "holdoutPct": 19.99563,
            "validationLevel": null
        },
        "recommender": {
            "recommenderItemId": null,
            "isRecommender": false,
            "recommenderUserId": null
        },
        "advancedOptions": {
            "blueprintThreshold": null,
            "responseCap": false,
            "seed": null,
            "weights": null,
            "smartDownsampled": false,
            "majorityDownsamplingRate": null,
            "offset": null,
            "exposure": null
        },
        "positiveClass": 1,
        "maxTrainPct": 64.0035,
        "holdoutUnlocked": false,
        "targetType": "Binary"
    }
"""


@pytest.fixture
def project_with_target_data(project_with_target_json):
    data = json.loads(project_with_target_json)
    return from_api(data)


@pytest.fixture
def smart_sampled_project_server_data(project_with_target_json):
    base_data = json.loads(project_with_target_json)
    return dict(base_data, smartDownsampled=True, majorityDownsamplingRate=50.5)


@pytest.fixture
def project(project_with_target_data):
    return Project.from_data(project_with_target_data)


@pytest.fixture
def smart_sampled_project(smart_sampled_project_data):
    return Project.from_data(smart_sampled_project_data)


@pytest.fixture
def project_without_target(project_without_target_data):
    return Project.from_data(project_without_target_data)


@pytest.fixture
def async_failure_json():
    return """
    {
        "status": "ERROR",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def async_aborted_json():
    return """
    {
        "status": "ABORTED",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def async_running_json():
    return """
    {
        "status": "RUNNING",
        "message": "",
        "code": 0,
        "created": "2016-07-22T12:00:00.123456Z"
    }
    """


@pytest.fixture
def model_id():
    """
    The id of the model used in the fixtures

    Returns
    -------
    model_id : str
        The id of the model used in the fixtures
    """
    return '556ce973100d2b6e51ca9657'


@pytest.fixture
def model_json():
    return """
{
    "featurelistId": "57993241bc92b715ed0239ee",
    "processes": [
      "One",
      "Two",
      "Three"
    ],
    "featurelistName": "Informative Features",
    "projectId": "556cdfbb100d2b0e88585195",
    "samplePct": 64,
    "modelType": "Gradient Boosted Trees Classifier",
    "metrics": {
      "AUC": {
        "holdout": null,
        "validation": 0.73376,
        "crossValidation": null
      },
      "Rate@Top5%": {
        "holdout": null,
        "validation": 0.44218,
        "crossValidation": null
      },
      "Rate@TopTenth%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "RMSE": {
        "holdout": null,
        "validation": 0.27966,
        "crossValidation": null
      },
      "LogLoss": {
        "holdout": null,
        "validation": 0.2805,
        "crossValidation": null
      },
      "FVE Binomial": {
        "holdout": null,
        "validation": 0.12331,
        "crossValidation": null
      },
      "Gini Norm": {
        "holdout": null,
        "validation": 0.46752,
        "crossValidation": null
      },
      "Rate@Top10%": {
        "holdout": null,
        "validation": 0.34812,
        "crossValidation": null
      }
    },
    "modelCategory": "model",
    "isFrozen": false,
    "blueprintId": "de628edee06f2b23218767a245e45ae1",
    "id": "556ce973100d2b6e51ca9657"
  }
    """


@pytest.fixture
def blender_json(model_json):
    data = json.loads(model_json)
    data['modelIds'] = ['556ce973100d2b6e51ca9656', '556ce973100d2b6e51ca9655']
    data['blenderMethod'] = 'AVG'
    return json.dumps(data)


@pytest.fixture
def blenders_list_response_json(blender_json):
    data = json.loads(blender_json)
    return json.dumps({'data': [data, data]})


@pytest.fixture
def frozen_json(model_json):
    data = json.loads(model_json)
    data['parentModelId'] = '556ce973100e2b6e51ca8657'
    return json.dumps(data)


@pytest.fixture
def frozen_models_list_response(frozen_json):
    data = json.loads(frozen_json)
    return {'data': [data, data], 'count': 2, 'prev': None, 'next': None}


@pytest.fixture
def model_data(model_json):
    return from_api(json.loads(model_json))


@pytest.fixture
def datetime_model_data(model_data):
    modified_model_data = dict(model_data)
    modified_model_data.pop('sample_pct')
    modified_model_data['training_start_date'] = '2015-12-10T19:00:00.000000Z'
    modified_model_data['training_end_date'] = '2016-12-10T19:00:00.000000Z'
    return modified_model_data


@pytest.fixture
def one_model(model_data):
    return Model.from_data(model_data)


@pytest.fixture
def blender_data(blender_json):
    return from_api(json.loads(blender_json))


@pytest.fixture
def one_blender(blender_data):
    return BlenderModel.from_data(blender_data)


@pytest.fixture
def frozen_data(frozen_json):
    return from_api(json.loads(frozen_json))


@pytest.fixture
def one_frozen(frozen_data):
    return FrozenModel.from_data(frozen_data)


@pytest.fixture
def prime_model_id():
    return '57aa68e1ccf94e1ce3197743'


@pytest.fixture
def prime_model_json():
    return """
    {
    "featurelistId": "57aa1c46ccf94e5bb073841b",
    "processes": [],
    "featurelistName": "Informative Features",
    "projectId": "556cdfbb100d2b0e88585195",
    "samplePct": 63.863,
    "modelType": "DataRobot Prime",
    "metrics": {
      "AUC": {
        "holdout": null,
        "validation": 0.8559,
        "crossValidation": null
      },
      "Rate@Top5%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "Rate@TopTenth%": {
        "holdout": null,
        "validation": 1,
        "crossValidation": null
      },
      "RMSE": {
        "holdout": null,
        "validation": 0.37973,
        "crossValidation": null
      },
      "LogLoss": {
        "holdout": null,
        "validation": 0.41848,
        "crossValidation": null
      },
      "FVE Binomial": {
        "holdout": null,
        "validation": 0.32202,
        "crossValidation": null
      },
      "Gini Norm": {
        "holdout": null,
        "validation": 0.7118,
        "crossValidation": null
      },
      "Rate@Top10%": {
        "holdout": null,
        "validation": 0.66667,
        "crossValidation": null
      }
    },
    "modelCategory": "prime",
    "blueprintId": "bcfb575932da72a92d01837a6c42a36f5cc56cbdab7d92f43b88e114179f2942",
    "id": "57aa68e1ccf94e1ce3197743",
    "rulesetId": 3,
    "score": 0.41847989771503824,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 323
    }
    """


@pytest.fixture
def prime_model_server_data(prime_model_json):
    return json.loads(prime_model_json)


@pytest.fixture
def prime_data(prime_model_server_data):
    return from_api(prime_model_server_data)


@pytest.fixture
def prime_model(prime_model_server_data):
    return PrimeModel.from_server_data(prime_model_server_data)


@pytest.fixture
def ruleset_with_model_json():
    return """
    {
    "projectId": "556cdfbb100d2b0e88585195",
    "rulesetId": 3,
    "score": 0.41847989771503824,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 323,
    "modelId": "57aa68e1ccf94e1ce3197743"
    }
    """


@pytest.fixture
def ruleset_without_model_json():
    return """
    {
    "projectId": "556cdfbb100d2b0e88585195",
    "rulesetId": 2,
    "score": 0.428702,
    "parentModelId": "556ce973100d2b6e51ca9657",
    "ruleCount": 161,
    "modelId": null
    }
    """


@pytest.fixture
def ruleset_with_model_server_data(ruleset_with_model_json):
    return json.loads(ruleset_with_model_json)


@pytest.fixture
def ruleset_without_model_server_data(ruleset_without_model_json):
    return json.loads(ruleset_without_model_json)


@pytest.fixture
def ruleset_with_model(ruleset_with_model_server_data):
    return Ruleset.from_server_data(ruleset_with_model_server_data)


@pytest.fixture
def ruleset_without_model(ruleset_without_model_server_data):
    return Ruleset.from_server_data(ruleset_without_model_server_data)


@pytest.fixture
def prime_file_json():
    return """
    {
    "id": "57fa1c41ccf94e59a9024e87",
    "projectId": "556cdfbb100d2b0e88585195",
    "parentModelId": "556ce973100d2b6e51ca9657",
    "modelId": "57aa68e1ccf94e1ce3197743",
    "rulesetId": 3,
    "language": "Python",
    "isValid": true
    }
    """


@pytest.fixture
def prime_file_server_data(prime_file_json):
    return json.loads(prime_file_json)


@pytest.fixture
def prime_file(prime_file_server_data):
    return PrimeFile.from_server_data(prime_file_server_data)


@pytest.fixture
def job_id():
    return '13'


@pytest.fixture
def job_url(project_id, job_id):
    return 'https://host_name.com/projects/{}/jobs/{}/'.format(project_id, job_id)


@pytest.fixture
def base_job_server_data(job_id, project_id, job_url):
    return {
        'status': None,
        'url': job_url,
        'id': job_id,
        'jobType': None,
        'projectId': project_id
    }


@pytest.fixture
def base_job_running_server_data(base_job_server_data):
    return dict(base_job_server_data, status=enums.QUEUE_STATUS.INPROGRESS)


@pytest.fixture
def base_job_completed_server_data(base_job_server_data):
    return dict(base_job_server_data, status=enums.QUEUE_STATUS.COMPLETED)


@pytest.fixture
def prime_model_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_MODEL)


@pytest.fixture
def prime_model_job_completed_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_MODEL)


@pytest.fixture
def prime_model_job_creation_response(project_url, job_url):
    responses.add(responses.POST, '{}primeModels/'.format(project_url), body='', status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def prime_model_job_completed_response(prime_model_job_completed_server_data,
                                       job_url, project_url, prime_model_id,
                                       prime_model_server_data):
    prime_model_url = '{}primeModels/{}/'.format(project_url, prime_model_id)
    responses.add(responses.GET, job_url, body=json.dumps(prime_model_job_completed_server_data),
                  status=303, adding_headers={'Location': prime_model_url},
                  content_type='application/json')
    responses.add(responses.GET, prime_model_url, body=json.dumps(prime_model_server_data),
                  content_type='application/json')


@pytest.fixture
def frozen_model_job_completed_server_data(base_job_completed_server_data, one_model):
    return dict(base_job_completed_server_data,
                jobType=enums.JOB_TYPE.MODEL,
                modelType=one_model.model_type,
                processes=one_model.processes,
                blueprintId=one_model.blueprint_id)


@pytest.fixture
def rci_json():
    """ ReasonCodesInitialization GET json
    """
    return """{
      "projectId": "556cdfbb100d2b0e88585195",
      "reasonCodesSample": [
        {
          "reasonCodes": [
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 115.59591064453161,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": "Wheel Loader - 250.0 to 275.0 Horsepower",
              "label": "YearMade",
              "strength": 60.62447509765593,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 54.968029785156205,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1.0,
              "label": "YearMade",
              "strength": 51.503918457031205,
              "feature": "auctioneerID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1763636,
              "label": "YearMade",
              "strength": 45.668481445312636,
              "feature": "SalesID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "-2",
              "label": "YearMade",
              "strength": 35.2052001953125,
              "feature": "fiModelSeries",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Utah",
              "label": "YearMade",
              "strength": 25.33120117187491,
              "feature": "state",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "WA450",
              "label": "YearMade",
              "strength": 21.51729736328116,
              "feature": "fiBaseModel",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": 383693,
              "label": "YearMade",
              "strength": 17.533557128906068,
              "feature": "MachineID",
              "qualitativeStrength": "+"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 12.142431640625091,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "+"
            }
          ],
          "prediction": 2059.988037109375,
          "rowId": 7459,
          "predictionValues": [{"value": 2059.988037109375, "label": "YearMade"}]
        },
        {
          "reasonCodes": [
            {
              "featureValue": "50DZTS",
              "label": "YearMade",
              "strength": 86.08530273437486,
              "feature": "fiModelDesc",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 2063.0,
              "label": "YearMade",
              "strength": 49.44567871093727,
              "feature": "MachineHoursCurrentMeter",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Medium",
              "label": "YearMade",
              "strength": 43.51775771484381,
              "feature": "UsageBand",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "Hydraulic Excavator, Track - 4.0 to 5.0 Metric Tons",
              "label": "YearMade",
              "strength": 33.40051269531273,
              "feature": "fiProductClassDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 1250726400,
              "label": "YearMade",
              "strength": 30.625183105468295,
              "feature": "saledate",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "",
              "label": "YearMade",
              "strength": 26.95623779296875,
              "feature": "Stick_Length",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 21500,
              "label": "YearMade",
              "strength": 25.61611328124968,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": 22117,
              "label": "YearMade",
              "strength": 25.539965820312545,
              "feature": "ModelID",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "D",
              "label": "YearMade",
              "strength": 21.061962890624955,
              "feature": "fiSecondaryDesc",
              "qualitativeStrength": "++"
            },
            {
              "featureValue": "50",
              "label": "YearMade",
              "strength": 20.641674804687682,
              "feature": "fiBaseModel",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 2056.3349609375,
          "rowId": 6934,
          "predictionValues": [{"value": 2056.3349609375, "label": "YearMade"}]
        }
      ],
      "modelId": "578e59a41ced2e5a9eb18965"
    }"""


@pytest.fixture
def rci_data(rci_json):
    return from_api(json.loads(rci_json))


@pytest.fixture
def one_rci(rci_data):
    return ReasonCodesInitialization.from_data(rci_data)


@pytest.fixture
def rc_list_json():
    """List of reason codes"""
    return """{
      "count": 3,
      "next": null,
      "data": [
        {
          "finishTime": 1482414443.292265,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..7/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 2,
          "id": "585bd862dd56a7482bdcb567",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482414817.200652,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..6/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 3,
          "id": "585bda05dd56a7483781c4b6",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        },
        {
          "finishTime": 1482418494.81697,
          "numColumns": 50,
          "reasonCodesLocation": "https://host_name.com/projects/5..4/reasonCodes/5..a/",
          "thresholdHigh": 0.835627812,
          "projectId": "578e565a218ef11f13fd2f64",
          "thresholdLow": 0.1,
          "maxCodes": 3,
          "id": "585be85cdd56a7482bdcb56a",
          "datasetId": "585bbd1dc522f21821107d6e",
          "modelId": "578e59a41ced2e5a9eb18965"
        }
      ],
      "previous": null
    }"""


@pytest.fixture
def rc_json():
    """ ReasonCodes GET json
    """
    return """{
      "finishTime": 1482424475.33823,
      "numColumns": 50,
      "reasonCodesLocation": "https://host_name.com/projects/578e...2f64/reasonCodes/585b...9e0f/",
      "thresholdHigh": 2010.0490966796874,
      "projectId": "556cdfbb100d2b0e88585195",
      "thresholdLow": 1706.8041999316963,
      "maxCodes": 3,
      "id": "585ba071dd56a72ec7109e0f",
      "datasetId": "585b9d71c522f20d38107d6e",
      "modelId": "578e59a41ced2e5a9eb18960"
    }"""


@pytest.fixture
def rc_data(rc_json):
    return from_api(json.loads(rc_json))


@pytest.fixture
def one_rc(rc_data):
    return ReasonCodes.from_data(rc_data)


@pytest.fixture
def rcp_json():
    """ ReasonCodesPage GET json
    """
    return """{
      "count": 6,
      "reasonCodesRecordLocation": "https://host_name.com/projects/5..4/reasonCodesRecords/5..7/",
      "next": null,
      "data": [
        {
          "reasonCodes": [
            {
              "featureValue": 66000,
              "label": "YearMade",
              "strength": 113.18563232421889,
              "feature": "SalePrice",
              "qualitativeStrength": "+++"
            },
            {
              "featureValue": 1139246,
              "label": "YearMade",
              "strength": -84.37227783203139,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1843.3448486328125,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 1843.3448486328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 121,
              "label": "YearMade",
              "strength": -140.4251185424805,
              "feature": "datasource",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 57000,
              "label": "YearMade",
              "strength": 107.4373168945308,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1730.043701171875,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 1730.043701171875,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139249,
              "label": "YearMade",
              "strength": -201.79853515624995,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 10000,
              "label": "YearMade",
              "strength": -144.66733398437532,
              "feature": "SalePrice",
              "qualitativeStrength": "---"
            }
          ],
          "prediction": 1751.1256103515625,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 1751.1256103515625,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139251,
              "label": "YearMade",
              "strength": -222.99357910156255,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 1305763200,
              "label": "YearMade",
              "strength": -107.39191894531268,
              "feature": "saledate",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1663.2529296875,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 1663.2529296875,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139253,
              "label": "YearMade",
              "strength": -209.14246826171893,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 11000,
              "label": "YearMade",
              "strength": -67.03535156250018,
              "feature": "SalePrice",
              "qualitativeStrength": "--"
            }
          ],
          "prediction": 1724.86328125,
          "rowId": 4,
          "predictionValues": [
            {
              "value": 1724.86328125,
              "label": "YearMade"
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 1139255,
              "label": "YearMade",
              "strength": -81.31390380859398,
              "feature": "SalesID",
              "qualitativeStrength": "---"
            },
            {
              "featureValue": 26500,
              "label": "YearMade",
              "strength": 52.808850097656205,
              "feature": "SalePrice",
              "qualitativeStrength": "++"
            }
          ],
          "prediction": 1881.0067138671875,
          "rowId": 5,
          "predictionValues": [
            {
              "value": 1881.0067138671875,
              "label": "YearMade"
            }
          ]
        }
      ],
      "id": "585bd862dd56a7482bdcb567",
      "previous": null
    }"""


@pytest.fixture
def rcp_data(rcp_json):
    return from_api(json.loads(rcp_json))


@pytest.fixture
def rcp_json_classification():
    return """{
      "count": 4,
      "reasonCodesRecordLocation": "https://host_name.com/projects/5...2/reasonCodesRecords/5...c/",
      "next": null,
      "data": [
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 0,
          "predictionValues": [
            {
              "value": 0.2893982637455831,
              "label": 1
            },
            {
              "value": 0.7106017362544169,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 1,
          "predictionValues": [
            {
              "value": 0.29246609348399255,
              "label": 1
            },
            {
              "value": 0.7075339065160074,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [],
          "prediction": 0,
          "rowId": 2,
          "predictionValues": [
            {
              "value": 0.3840762929836601,
              "label": 1
            },
            {
              "value": 0.6159237070163399,
              "label": 0
            }
          ]
        },
        {
          "reasonCodes": [
            {
              "featureValue": 3,
              "strength": -0.3712657301791762,
              "feature": "number_diagnoses",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "UN",
              "strength": -0.3112189393477769,
              "feature": "payer_code",
              "qualitativeStrength": "--",
              "label": 1
            },
            {
              "featureValue": "41",
              "strength": -0.2607262089753333,
              "feature": "diag_2",
              "qualitativeStrength": "--",
              "label": 1
            }
          ],
          "prediction": 0,
          "rowId": 3,
          "predictionValues": [
            {
              "value": 0.12064479565777504,
              "label": 1
            },
            {
              "value": 0.8793552043422249,
              "label": 0
            }
          ]
        }
      ],
      "id": "58d1ef9fc80891088925890c",
      "previous": null
    }"""


@pytest.fixture
def rcp_data_classification(rcp_json_classification):
    return from_api(json.loads(rcp_json_classification))


@pytest.fixture
def one_rcp(rcp_data):
    with mock.patch('datarobot.models.reason_codes.t.URL.check_and_return',
                    side_effect=lambda url: url):
        return ReasonCodesPage.from_data(rcp_data)



import json

import mock
import pytest

import responses

from datarobot import Project
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.utils import to_api


class SpecialProject(Project):
    """
    Class used by TestProjectExtensibility.

    Note that this is very similar to AdminProject for MBTests.
    """
    _fields = Project._fields

    @classmethod
    def create(cls, sourcedata, project_name='New Project', max_wait=DEFAULT_MAX_WAIT, **kwargs):
        form_data = cls._construct_create_form_data(project_name=project_name)

        if kwargs:
            form_data.update(kwargs)
        return cls._create_project_with_form_data(
            sourcedata,
            form_data,
            max_wait=max_wait,
        )

    def _construct_aim_payload(self, *args, **kwargs):
        """
        Constructs the AIM payload that is POSTed to the server to set the target.
        Adds extras to the AIM request for special things.

        Returns
        -------
        dict
        """
        payload = super(SpecialProject, self)._construct_aim_payload(*args, **kwargs)
        payload.update(self._aim_extras)
        return payload

    def set_target(self, *args, **kwargs):
        """
        Sets the target.
        Adds extras to the AIM request.
        See Project.set_target for parameters and return value.
        """
        import inspect
        # gets the method arguments, excluding self (1st arg)
        args_list = inspect.getargspec(Project.set_target).args[1:]
        # subset the dict on only the keys that the method accepts
        set_target_args = {k: v for k, v in kwargs.items() if k in args_list}
        # extract the extra kwargs to inject into the AIM payload
        self._aim_extras = {k: v for k, v in kwargs.items() if k not in args_list}
        return super(SpecialProject, self).set_target(*args, **set_target_args)


@pytest.fixture
def extras():
    return {
        'special': 100,
        'experimental_magic': True
    }


@responses.activate
@pytest.mark.usefixtures('client', 'project_creation_responses')
def test_create_with_extras(extras,
                            async_url):
    """
    Creating a SpecialProject with extra kwargs should be appended to the POSTed create payload.
    """
    max_wait = 600

    with mock.patch.object(
            SpecialProject,
            'from_async',
            wraps=SpecialProject.from_async) as mock_from_async:
        special_project = SpecialProject.create(
            'https://example.com/data.csv',
            'A Project Name',
            max_wait,
            **extras)

    mock_from_async.assert_called_once_with(async_url, max_wait)
    assert special_project.project_name == 'A Project Name'
    assert isinstance(special_project, SpecialProject)

    posted_data = json.loads(responses.calls[0].request.body)
    for extra in to_api(extras).items():
        # note that to_api will convert snake_case to CamelCase on dict keys
        assert extra in posted_data.items()


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_with_extras_should_fail_on_normal_project(extras):
    """
    By default, a normal Project should prevent any extra kwargs to create method.
    """
    with pytest.raises(TypeError):
        Project.create(
            'https://example.com/data.csv',
            'Project Name',
            600,
            **extras)


@responses.activate
@pytest.mark.usefixtures('client')
def test_set_target_with_extras_should_fail_on_normal_project(project_without_target):
    """
    By default, a normal Project should prevent any extra kwargs to set_target.
    """
    extras = {
        'special': 100,
        'experimental_magic': True
    }

    with pytest.raises(TypeError):
        project_without_target.set_target('Bullseye', **extras)


@responses.activate
@pytest.mark.usefixtures('client', 'project_aim_responses')
def test_set_target_with_extras(extras,
                                project_without_target_data):
    """
    With a SpecialProject that subclasses Project, it should be possible to pass additional
    kwargs to set_target as part of the POSTed payload to the server.
    """
    special_project = SpecialProject.from_data(project_without_target_data)
    special_project.set_target('Bullseye', **extras)
    posted_data = json.loads(responses.calls[0].request.body)
    for extra in to_api(extras).items():
        # note that to_api will convert snake_case to CamelCase on dict keys
        assert extra in posted_data.items()



import trafaret as t

from datarobot.models.api_object import APIObject
from ..utils import deprecation_warning, encode_utf8_if_py2


class BlueprintTaskDocument(APIObject):
    """ Document describing a task from a blueprint.

    Attributes
    ----------
    title : str
        Title of document.
    task : str
        Name of the task described in document.
    description : str
        Task description.
    parameters : list of dict(name, type, description)
        Parameters that task can receive in human-readable format.
    links : list of dict(name, url)
        External links used in document
    references : list of dict(name, url)
        References used in document. When no link available url equals None.
    """
    _converter = t.Dict({
        t.Key('title'): t.String,
        t.Key('task'): t.String(allow_blank=True),
        t.Key('description'): t.String(allow_blank=True),
        t.Key('parameters'): t.List(t.Dict({
            t.Key('name'): t.String,
            t.Key('type'): t.String,
            t.Key('description'): t.String,
        }).ignore_extra('*')),
        t.Key('links'): t.List(t.Dict({
            t.Key('name'): t.String,
            t.Key('url'): t.String,
        }).ignore_extra('*')),
        t.Key('references'): t.List(t.Dict({
            t.Key('name'): t.String,
            # from_api method drops None, so we need this when there is no url
            t.Key('url', optional=True, default=None): t.String | t.Null,
        }).ignore_extra('*')),
    }).ignore_extra('*')

    def __init__(self, title=None, task=None,
                 description=None, parameters=None,
                 links=None, references=None):
        self.title = title
        self.task = task
        self.description = description
        self.parameters = parameters
        self.links = links
        self.references = references

    def __repr__(self):
        return encode_utf8_if_py2(u'BlueprintTaskDocument({})'.format(self.title))


class BlueprintChart(APIObject):
    """ A Blueprint chart that can be used to understand data flow in blueprint.

    Attributes
    ----------
    nodes : list of dict (id, label)
        Chart nodes, id unique in chart.
    edges : list of tuple (id1, id2)
        Directions of data flow between blueprint chart nodes.
    """

    _converter = t.Dict({
        t.Key('nodes', optional=True): t.List(t.Dict({
            t.Key('id'): t.String,
            t.Key('label'): t.String
        })),
        t.Key('edges', optional=True): t.List(t.Tuple(t.String, t.String)),
    })

    def __init__(self, nodes, edges):
        self.nodes = nodes
        self.edges = edges

    def __repr__(self):
        return encode_utf8_if_py2(u'BlueprintChart({} nodes, {} edges)'.format(len(self.nodes),
                                                                               len(self.edges)))

    @classmethod
    def get(cls, project_id, blueprint_id):
        """ Retrieve a blueprint chart.

        Parameters
        ----------
        project_id : str
            The project's id.
        blueprint_id : str
            Id of blueprint to retrieve chart.

        Returns
        -------
        BlueprintChart
            The queried blueprint chart.
        """
        url = 'projects/{}/blueprints/{}/blueprintChart/'.format(project_id, blueprint_id)
        return cls.from_location(url)

    def to_graphviz(self):
        """ Get blueprint chart in graphviz DOT format.

        Returns
        -------
        unicode
            String representation of chart in graphviz DOT language.
        """
        digraph = u'digraph "Blueprint Chart" {'
        digraph += u'\ngraph [rankdir=LR]'
        for node in self.nodes:
            digraph += u'\n{id} [label="{label}"]'.format(id=node['id'], label=node['label'])
        for edge in self.edges:
            digraph += u'\n{id0} -> {id1}'.format(id0=edge[0], id1=edge[1])
        digraph += u'\n}'
        return digraph


class Blueprint(APIObject):
    """ A Blueprint which can be used to fit models

    Attributes
    ----------
    id : str
        the id of the blueprint
    processes : list of str
        the processes used by the blueprint
    model_type : str
        the model produced by the blueprint
    project_id : str
        the project the blueprint belongs to
    blueprint_category : str
        (New in version v2.6) Describes the category of the blueprint and the kind of model it
        produces.
    """
    _converter = t.Dict({
        t.Key('id', optional=True): t.String(),
        t.Key('processes', optional=True): t.List(t.String()),
        t.Key('model_type', optional=True): t.String(),
        t.Key('project_id', optional=True): t.String(),
        t.Key('blueprint_category', optional=True): t.String()
    }).allow_extra('*')

    def __init__(self, id=None, processes=None, model_type=None, project_id=None,
                 blueprint_category=None):
        if isinstance(id, dict):
            deprecation_warning('Blueprint instantiation from a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Blueprint.from_data instead')
            self.__init__(**id)
        else:
            self.id = id
            self.processes = processes
            self.model_type = model_type
            self.project_id = project_id
            self.blueprint_category = blueprint_category

    def __repr__(self):
        return encode_utf8_if_py2(u'Blueprint({})'.format(self.model_type))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def get(cls, project_id, blueprint_id):
        """ Retrieve a blueprint.

        Parameters
        ----------
        project_id : str
            The project's id.
        blueprint_id : str
            Id of blueprint to retrieve.

        Returns
        -------
        blueprint : Blueprint
            The queried blueprint.
        """
        url = 'projects/{}/blueprints/{}/'.format(project_id, blueprint_id)
        return cls.from_location(url)

    def get_chart(self):
        """ Retrieve a chart.

        Returns
        -------
        BlueprintChart
            The current blueprint chart.
        """
        return BlueprintChart.get(self.project_id, self.id)

    def get_documents(self):
        """ Get documentation for tasks used in the blueprint.

        Returns
        -------
        list of BlueprintTaskDocument
            All documents available for blueprint.
        """
        url = 'projects/{}/blueprints/{}/blueprintDocs/'.format(self.project_id, self.id)
        return [BlueprintTaskDocument.from_server_data(data) for data in self._server_data(url)]



import datetime
import json

import pytest
import responses

from datarobot import BacktestSpecification, DatetimePartitioning, DatetimePartitioningSpecification
from datarobot import partitioning_methods


def test_future_proof(datetime_partition_server_data):
    future_response = dict(datetime_partition_server_data, future='newKey')
    DatetimePartitioning.from_server_data(future_response)


def test_construct_duration_string():
    duration = partitioning_methods.construct_duration_string(years=1, months=2, days=3,
                                                              hours=4, minutes=5, seconds=6)
    assert duration == 'P1Y2M3DT4H5M6S'


def test_construct_duration_string_empty():
    assert partitioning_methods.construct_duration_string() == 'P0Y0M0DT0H0M0S'


def test_collect_payload_fails_cleanly_on_type_error(datetime_partition_spec):
    datetime_partition_spec.holdout_start_date = str(datetime_partition_spec.holdout_start_date)
    with pytest.raises(ValueError) as exc_info:
        datetime_partition_spec.collect_payload()
    assert 'expected holdout_start_date to be a datetime.datetime' in str(exc_info.value)


def test_backtest_to_specification(datetime_partition):
    backtest = datetime_partition.backtests[0]
    backtest_spec = backtest.to_specification()
    assert backtest_spec.index == backtest.index
    assert backtest_spec.gap_duration == backtest.gap_duration
    assert backtest_spec.validation_start_date
    assert backtest_spec.validation_duration == backtest.validation_duration


def test_partition_to_specification(datetime_partition):
    part_spec = datetime_partition.to_specification()
    assert part_spec.datetime_partition_column == datetime_partition.datetime_partition_column
    assert (part_spec.autopilot_data_selection_method ==
            datetime_partition.autopilot_data_selection_method)
    assert part_spec.validation_duration == datetime_partition.validation_duration
    assert part_spec.holdout_start_date == datetime_partition.holdout_start_date
    assert part_spec.holdout_duration == datetime_partition.holdout_duration
    assert part_spec.gap_duration == datetime_partition.gap_duration
    assert part_spec.number_of_backtests == datetime_partition.number_of_backtests
    assert len(part_spec.backtests) == len(datetime_partition.backtests)
    for bt_spec in part_spec.backtests:
        assert isinstance(bt_spec, BacktestSpecification)


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve(project_id, project_url, datetime_partition_after_target_server_data):
    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_after_target_server_data)

    partition = DatetimePartitioning.get(project_id)

    assert partition.primary_training_row_count == 100
    assert partition.available_training_row_count == 100
    assert partition.gap_row_count == 100
    assert partition.holdout_row_count == 100
    assert partition.total_row_count == 100

    backtest = partition.backtests[0]
    assert backtest.primary_training_row_count == 100
    assert backtest.available_training_row_count == 100
    assert backtest.gap_row_count == 100
    assert backtest.validation_row_count == 100
    assert backtest.total_row_count == 100


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate(project_id, project_url, datetime_partition_server_data,
                  holdout_start_date, datetime_partition_spec):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data)

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)

    # test that we send the right things
    actual_payload = json.loads(responses.calls[0].request.body)
    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in datetime_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': datetime_partition_spec.datetime_partition_column,
        'autopilotDataSelectionMethod': datetime_partition_spec.autopilot_data_selection_method,
        'validationDuration': datetime_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': datetime_partition_spec.holdout_duration,
        'gapDuration': datetime_partition_spec.gap_duration,
        'numberOfBacktests': datetime_partition_spec.number_of_backtests,
        'backtests': expected_backtests
    }
    assert actual_payload == expected_payload

    # test that we get the right thing back
    assert partition.project_id == project_id
    assert (partition.datetime_partition_column ==
            datetime_partition_server_data['datetimePartitionColumn'])
    assert partition.date_format == datetime_partition_server_data['dateFormat']
    assert (partition.autopilot_data_selection_method ==
            datetime_partition_server_data['autopilotDataSelectionMethod'])
    assert partition.validation_duration == datetime_partition_server_data['validationDuration']

    assert isinstance(partition.available_training_start_date, datetime.datetime)
    assert (partition.available_training_duration ==
            datetime_partition_server_data['availableTrainingDuration'])
    assert isinstance(partition.available_training_end_date, datetime.datetime)

    assert isinstance(partition.primary_training_start_date, datetime.datetime)
    assert (partition.primary_training_duration ==
            datetime_partition_server_data['primaryTrainingDuration'])
    assert isinstance(partition.primary_training_end_date, datetime.datetime)

    assert isinstance(partition.gap_start_date, datetime.datetime)
    assert partition.gap_duration == datetime_partition_server_data['gapDuration']
    assert isinstance(partition.gap_end_date, datetime.datetime)

    assert partition.holdout_start_date == holdout_start_date
    assert partition.holdout_duration == datetime_partition_server_data['holdoutDuration']
    assert isinstance(partition.holdout_end_date, datetime.datetime)

    assert partition.number_of_backtests == datetime_partition_server_data['numberOfBacktests']
    [backtest] = partition.backtests
    [backtest_data] = datetime_partition_server_data['backtests']
    assert backtest.index == backtest_data['index']
    assert isinstance(backtest.available_training_start_date, datetime.datetime)
    assert backtest.available_training_duration == backtest_data['availableTrainingDuration']
    assert isinstance(backtest.available_training_end_date, datetime.datetime)
    assert isinstance(backtest.primary_training_start_date, datetime.datetime)
    assert backtest.primary_training_duration == backtest_data['primaryTrainingDuration']
    assert isinstance(backtest.primary_training_end_date, datetime.datetime)
    assert isinstance(backtest.gap_start_date, datetime.datetime)
    assert backtest.gap_duration == backtest_data['gapDuration']
    assert isinstance(backtest.gap_end_date, datetime.datetime)
    assert isinstance(backtest.validation_start_date, datetime.datetime)
    assert backtest.validation_duration == backtest_data['validationDuration']
    assert isinstance(backtest.validation_end_date, datetime.datetime)


@responses.activate
@pytest.mark.usefixtures('client')
def test_from_minimum_specification(project_id, project_url, datetime_partition_server_data):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data)

    min_spec = DatetimePartitioningSpecification(
        datetime_partition_server_data['datetimePartitionColumn'])
    DatetimePartitioning.generate(project_id, min_spec)

    expected_data = {'datetimePartitionColumn':
                     datetime_partition_server_data['datetimePartitionColumn']}
    assert responses.calls[0].request.body == json.dumps(expected_data)


def test_to_dataframe(datetime_partition):
    df = datetime_partition.to_dataframe()
    expected_shape = (4*(1+datetime_partition.number_of_backtests), 3)
    assert df.shape == expected_shape



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import re
from datetime import datetime, date

import pandas as pd
import six
from dateutil import parser, tz

from .deprecation import deprecated, deprecation_warning  # noqa
from .sourcedata import is_urlsource, recognize_sourcedata, dataframe_to_buffer  # noqa

ALL_CAPITAL = re.compile(r'(.)([A-Z][a-z]+)')
CASE_SWITCH = re.compile(r'([a-z0-9])([A-Z])')
UNDERSCORES = re.compile(r'[a-z]_[a-z]{0,1}')


def underscorize(value):
    partial_result = ALL_CAPITAL.sub(r'\1_\2', value)
    return CASE_SWITCH.sub(r'\1_\2', partial_result).lower()


def underscoreToCamel(match):
    groups = match.group()
    if len(groups) == 2:
        # underscoreToCamel('sample_pct__gte') -> 'samplePct__gte'
        return groups
    return groups[0] + groups[2].upper()


def camelize(value):
    return UNDERSCORES.sub(underscoreToCamel, value)


def from_api(data, do_recursive=True):
    if type(data) not in (dict, list):
        return data
    if isinstance(data, list):
        return _from_api_list(data, do_recursive=do_recursive)
    return _from_api_dict(data, do_recursive=do_recursive)


def _from_api_dict(data, do_recursive=True):
    app_data = {}
    for k, v in six.iteritems(data):
        if v is None:
            continue
        app_data[underscorize(k)] = from_api(v, do_recursive=do_recursive) if do_recursive else v
    return app_data


def _from_api_list(data, do_recursive=True):
    return [from_api(datum, do_recursive=do_recursive) for datum in data]


def remove_empty_keys(metadata):
    return {k: v for k, v in metadata.items() if v is not None}


def parse_time(time_str):
    try:
        return parser.parse(time_str, tzinfos={'UTC': tz.tzutc()})
    except Exception:
        return time_str


def datetime_to_string(datetime_obj):
    """ Converts to isoformat after dropping microseconds
    """
    if not isinstance(datetime_obj, datetime):
        msg = 'expected to be passed a datetime.datetime, was passed {}'.format(type(datetime_obj))
        raise ValueError(msg)
    stripped = datetime_obj.replace(microsecond=0)
    return stripped.isoformat()


def to_api(data):
    """
    :param data: dictionary {'max_digits': 1}
    :return: {'maxDigits': 1}
    """
    if not data:
        return {}
    assert isinstance(data, dict), 'Wrong type'
    return _to_api_item(data)


def _to_api_item(item):
    if isinstance(item, dict):
        dense_item = remove_empty_keys(item)
        return {camelize(k): _to_api_item(v) for k, v in six.iteritems(dense_item)}
    elif isinstance(item, list):
        return [_to_api_item(subitem) for subitem in item]
    elif isinstance(item, datetime):
        return datetime_to_string(item)
    elif isinstance(item, date):
        return item.isoformat()
    else:
        return item


def get_id_from_response(response):
    location_string = response.headers['Location']
    return get_id_from_location(location_string)


def get_id_from_location(location_string):
    return location_string.split('/')[-2]


def get_duplicate_features(features):
    duplicate_features = set()
    seen_features = set()
    for feature in features:
        if feature in seen_features:
            duplicate_features.add(feature)
        else:
            seen_features.add(feature)
    return list(duplicate_features)


def raw_prediction_response_to_dataframe(pred_response):
    """Raw predictions for classification come as nested json objects.

    This will extract it to be a single dataframe.

    Parameters
    ----------
    pred_response : dict
        The loaded json object returned from the prediction route.

    Returns
    -------
    frame : pandas.DataFrame

    """
    predictions = from_api(pred_response)['predictions']
    frame = pd.DataFrame.from_records(predictions)
    return frame


def encode_utf8_if_py2(string):
    """__repr__ is supposed to return string (bytes) in 2 and string (unicode) in 3
    this function can be used to convert our unicode to strings in Python 2 but leave them alone
    in Python 3"""
    return string.encode('utf-8') if six.PY2 else string



import pytest
import responses

from datarobot import Project, FrozenModel


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_frozen_models_returns_valid_objects(frozen_models_list_response):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/frozenModels/',
                  json=frozen_models_list_response,
                  status=200)
    frozen_models = list(Project('p-id').get_frozen_models())
    assert len(frozen_models) == 2
    assert all(isinstance(frozen, FrozenModel) for frozen in frozen_models)



import abc
from datetime import datetime

import pandas as pd
import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.enums import CV_METHOD
from datarobot.utils import from_api, parse_time

__all__ = (
    'RandomCV',
    'StratifiedCV',
    'GroupCV',
    'UserCV',
    'RandomTVH',
    'UserTVH',
    'StratifiedTVH',
    'GroupTVH',
    'DatetimePartitioning',
    'DatetimePartitioningSpecification',
    'BacktestSpecification'
)


def get_class(cv_method, validation_type):
    if cv_method == CV_METHOD.DATETIME:
        raise ValueError('Cannot get_class for {} - use DatetimePartitioning.preview instead')
    classes = {
        'CV': {
            CV_METHOD.RANDOM: RandomCV,
            CV_METHOD.STRATIFIED: StratifiedCV,
            CV_METHOD.USER: UserCV,
            CV_METHOD.GROUP: GroupCV,
        },
        'TVH': {
            CV_METHOD.RANDOM: RandomTVH,
            CV_METHOD.STRATIFIED: StratifiedTVH,
            CV_METHOD.USER: UserTVH,
            CV_METHOD.GROUP: GroupTVH,
        },
    }
    try:
        return classes[validation_type][cv_method]
    except KeyError:
        err_msg = 'Error in getting class for cv_method={} and validation_type={}'
        raise ValueError(err_msg.format(cv_method, validation_type))


class PartitioningMethod(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def collect_payload(self):
        """ Set up the dict that should be sent to the server when setting the target
        Returns
        -------
        partitioning_spec : dict
        """
        return {}


class BasePartitioningMethod(PartitioningMethod):

    """This is base class to describe partitioning method
    with options"""

    cv_method = None
    validation_type = None
    seed = 0
    _data = None
    _static_fields = frozenset(['cv_method', 'validation_type'])

    def __init__(self, cv_method, validation_type, seed=0):
        self.cv_method = cv_method
        self.validation_type = validation_type
        self.seed = seed

    def collect_payload(self):
        keys = ('cv_method', 'validation_type', 'reps', 'user_partition_col',
                'training_level', 'validation_level', 'holdout_level', 'cv_holdout_level',
                'seed', 'validation_pct', 'holdout_pct', 'partition_key_cols')
        if not self._data:
            self._data = {key: getattr(self, key, None) for key in keys}
        return self._data

    def __repr__(self):
        if self._data:
            payload = {k: v for k, v in self._data.items()
                       if v is not None and k not in self._static_fields}
        else:
            self.collect_payload()
            return repr(self)
        return '{}({})'.format(self.__class__.__name__, payload)

    @classmethod
    def from_data(cls, data):
        """Can be used to instantiate the correct class of partitioning class
        based on the data
        """
        if data is None:
            return None
        cv_method = data.get('cv_method')
        validation_type = data.get('validation_type')
        other_params = {key: value for key, value in data.items()
                        if key not in ['cv_method', 'validation_type']}
        return get_class(cv_method, validation_type)(**other_params)


class BaseCrossValidation(BasePartitioningMethod):
    cv_method = None
    validation_type = 'CV'

    def __init__(self, cv_method, validation_type='CV'):
        self.cv_method = cv_method  # pragma: no cover
        self.validation_type = validation_type  # pragma: no cover


class BaseTVH(BasePartitioningMethod):
    cv_method = None
    validation_type = 'TVH'

    def __init__(self, cv_method, validation_type='TVH'):
        self.cv_method = cv_method  # pragma: no cover
        self.validation_type = validation_type  # pragma: no cover


class RandomCV(BaseCrossValidation):
    """A partition in which observations are randomly assigned to cross-validation groups
    and the holdout set.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    seed : int
        a seed to use for randomization
    """
    cv_method = 'random'

    def __init__(self, holdout_pct, reps, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.seed = seed  # pragma: no cover


class StratifiedCV(BaseCrossValidation):
    """A partition in which observations are randomly assigned to cross-validation groups
    and the holdout set, preserving in each group the same ratio of positive to negative cases as in
    the original data.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    seed : int
        a seed to use for randomization
    """
    cv_method = 'stratified'

    def __init__(self, holdout_pct, reps, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.seed = seed  # pragma: no cover


class GroupCV(BaseCrossValidation):
    """ A partition in which one column is specified, and rows sharing a common value
    for that column are guaranteed to stay together in the partitioning into cross-validation
    groups and the holdout set.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    reps : int
        number of cross validation folds to use
    partition_key_cols : list
        a list containing a single string, where the string is the name of the column whose
        values should remain together in partitioning
    seed : int
        a seed to use for randomization
    """
    cv_method = 'group'

    def __init__(self, holdout_pct, reps, partition_key_cols, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.reps = reps  # pragma: no cover
        self.partition_key_cols = partition_key_cols  # pragma: no cover
        self.seed = seed  # pragma: no cover


class UserCV(BaseCrossValidation):
    """ A partition where the cross-validation folds and the holdout set are specified by
    the user.

    Parameters
    ----------
    user_partition_col : string
        the name of the column containing the partition assignments
    cv_holdout_level
        the value of the partition column indicating a row is part of the holdout set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'user'

    def __init__(self, user_partition_col, cv_holdout_level, seed=0):
        self.user_partition_col = user_partition_col  # pragma: no cover
        self.cv_holdout_level = cv_holdout_level  # pragma: no cover
        self.seed = seed  # pragma: no cover


class RandomTVH(BaseTVH):
    """Specifies a partitioning method in which rows are randomly assigned to training, validation,
    and holdout.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'random'

    def __init__(self, holdout_pct, validation_pct, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.seed = seed  # pragma: no cover


class UserTVH(BaseTVH):
    """Specifies a partitioning method in which rows are assigned by the user to training,
    validation, and holdout sets.

    Parameters
    ----------
    user_partition_col : string
        the name of the column containing the partition assignments
    training_level
        the value of the partition column indicating a row is part of the training set
    validation_level
        the value of the partition column indicating a row is part of the validation set
    holdout_level
        the value of the partition column indicating a row is part of the holdout set (use
        None if you want no holdout set)
    seed : int
        a seed to use for randomization
    """
    cv_method = 'user'

    def __init__(self, user_partition_col, training_level, validation_level,
                 holdout_level, seed=0):
        self.user_partition_col = user_partition_col  # pragma: no cover
        self.training_level = training_level  # pragma: no cover
        self.validation_level = validation_level  # pragma: no cover
        self.holdout_level = holdout_level  # pragma: no cover
        self.seed = seed  # pragma: no cover


class StratifiedTVH(BaseTVH):
    """A partition in which observations are randomly assigned to train, validation, and
    holdout sets, preserving in each group the same ratio of positive to negative cases as in the
    original data.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    seed : int
        a seed to use for randomization
    """
    cv_method = 'stratified'

    def __init__(self, holdout_pct, validation_pct, seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.seed = seed  # pragma: no cover


class GroupTVH(BaseTVH):
    """A partition in which one column is specified, and rows sharing a common value
    for that column are guaranteed to stay together in the partitioning into the training,
    validation, and holdout sets.

    Parameters
    ----------
    holdout_pct : int
        the desired percentage of dataset to assign to holdout set
    validation_pct : int
        the desired percentage of dataset to assign to validation set
    partition_key_cols : list
        a list containing a single string, where the string is the name of the column whose
        values should remain together in partitioning
    seed : int
        a seed to use for randomization
    """
    cv_method = 'group'

    def __init__(self, holdout_pct, validation_pct, partition_key_cols,
                 seed=0):
        self.holdout_pct = holdout_pct  # pragma: no cover
        self.validation_pct = validation_pct  # pragma: no cover
        self.partition_key_cols = partition_key_cols  # pragma: no cover
        self.seed = seed  # pragma: no cover


def construct_duration_string(years=0, months=0, days=0, hours=0, minutes=0, seconds=0):
    """ Construct a valid string representing a duration in accordance with ISO8601

    A duration of six months, 3 days, and 12 hours could be represented as P6M3DT12H.

    Parameters
    ----------
    years : int
        the number of years in the duration
    months : int
        the number of months in the duration
    days : int
        the number of days in the duration
    hours : int
        the number of hours in the duration
    minutes : int
        the number of minutes in the duration
    seconds : int
        the number of seconds in the duration

    Returns
    -------
    duration_string: str
        The duration string, specified compatibly with ISO8601
    """
    return 'P{}Y{}M{}DT{}H{}M{}S'.format(years, months, days, hours, minutes, seconds)


_backtest_converter = t.Dict({
    t.Key('index'): t.Int(),
    t.Key('available_training_start_date'): parse_time,
    t.Key('available_training_duration'): t.String(),
    t.Key('available_training_row_count', optional=True): t.Int(),
    t.Key('available_training_end_date'): parse_time,
    t.Key('primary_training_start_date'): parse_time,
    t.Key('primary_training_duration'): t.String(),
    t.Key('primary_training_row_count', optional=True): t.Int(),
    t.Key('primary_training_end_date'): parse_time,
    t.Key('gap_start_date'): parse_time,
    t.Key('gap_duration'): t.String(),
    t.Key('gap_row_count', optional=True): t.Int(),
    t.Key('gap_end_date'): parse_time,
    t.Key('validation_start_date'): parse_time,
    t.Key('validation_duration'): t.String(),
    t.Key('validation_row_count', optional=True): t.Int(),
    t.Key('validation_end_date'): parse_time,
    t.Key('total_row_count', optional=True): t.Int()
}).ignore_extra('*')


class BacktestSpecification(object):
    """ Uniquely defines a Backtest used in a DatetimePartitioning

    Includes only the attributes of a backtest directly controllable by users.  The other attributes
    are assigned by the DataRobot application based on the project dataset and the user-controlled
    settings.

    All durations should be specified with a duration string such as those returned
    by the `partitioning_methods.construct_duration_string` helper method.

    Attributes
    ----------
    index : int
        the index of the backtest to update
    gap_duration : str
        the desired duration of the gap between training and validation scoring data for
        the backtest
    validation_start_date : datetime.datetime
        the desired start date of the validation scoring data for this backtest
    validation_duration : datetime.datetime
        the desired duration of the validation scoring data for this backtest
    """
    def __init__(self, index, gap_duration, validation_start_date, validation_duration):
        self.index = index
        self.gap_duration = gap_duration
        self.validation_start_date = validation_start_date
        self.validation_duration = validation_duration

    def collect_payload(self):
        if self.validation_start_date and not isinstance(self.validation_start_date, datetime):
            raise ValueError('expected validation_start_date to be a datetime.datetime')
        return {
            'index': self.index, 'gap_duration': self.gap_duration,
            'validation_start_date': self.validation_start_date,
            'validation_duration': self.validation_duration
        }


class Backtest(object):
    """ A backtest used to evaluate models trained in a datetime partitioned project

    When setting up a datetime partitioning project, backtests are specified by a
    BacktestSpecification.

    The available training data corresponds to all the data available for training, while the
    primary training data corresponds to the data that can be used to train while ensuring that all
    backtests are available.  If a model is trained with more data than is available in the primary
    training data, then all backtests may not have scores available.

    Attributes
    ----------
    index : int
        the index of the backtest
    available_training_start_date : datetime.datetime
        the start date of the available training data for this backtest
    available_training_duration : str
        the duration of available training data for this backtest
    available_training_row_count : int or None
        the number of rows of available training data for this backtest.  Only available when
        retrieving from a project where the target is set.
    available_training_end_date : datetime.datetime
        the end date of the available training data for this backtest
    primary_training_start_date : datetime.datetime
        the start date of the primary training data for this backtest
    primary_training_duration : str
        the duration of the primary training data for this backtest
    primary_training_row_count : int or None
        the number of rows of primary training data for this backtest.  Only available when
        retrieving from a project where the target is set.
    primary_training_end_date : datetime.datetime
        the end date of the primary training data for this backtest
    gap_start_date : datetime.datetime
        the start date of the gap between training and validation scoring data for this backtest
    gap_duration : str
        the duration of the gap between training and validation scoring data for this backtest
    gap_row_count : int or None
        the number of rows in the gap between training and validation scoring data for this
        backtest.  Only available when retrieving from a project where the target is set.
    gap_end_date : datetime.datetime
        the end date of the gap between training and validation scoring data for this backtest
    validation_start_date : datetime.datetime
        the start date of the validation scoring data for this backtest
    validation_duration : str
        the duration of the validation scoring data for this backtest
    validation_row_count : int or None
        the number of rows of validation scoring data for this backtest.  Only available when
        retrieving from a project where the target is set.
    validation_end_date : datetime.datetime
        the end date of the validation scoring data for this backtest
    total_row_count : int or None
        the number of rows in this backtest.  Only available when retrieving from a project where
        the target is set.
    """
    def __init__(self, index=None,
                 available_training_start_date=None, available_training_duration=None,
                 available_training_row_count=None, available_training_end_date=None,
                 primary_training_start_date=None, primary_training_duration=None,
                 primary_training_row_count=None, primary_training_end_date=None,
                 gap_start_date=None, gap_duration=None, gap_row_count=None, gap_end_date=None,
                 validation_start_date=None, validation_duration=None, validation_row_count=None,
                 validation_end_date=None, total_row_count=None):
        self.index = index
        self.available_training_start_date = available_training_start_date
        self.available_training_duration = available_training_duration
        self.available_training_row_count = available_training_row_count
        self.available_training_end_date = available_training_end_date
        self.primary_training_start_date = primary_training_start_date
        self.primary_training_duration = primary_training_duration
        self.primary_training_row_count = primary_training_row_count
        self.primary_training_end_date = primary_training_end_date
        self.gap_start_date = gap_start_date
        self.gap_duration = gap_duration
        self.gap_row_count = gap_row_count
        self.gap_end_date = gap_end_date
        self.validation_start_date = validation_start_date
        self.validation_duration = validation_duration
        self.validation_row_count = validation_row_count
        self.validation_end_date = validation_end_date
        self.total_row_count = total_row_count

    def to_specification(self):
        """ Render this backtest as a BacktestSpecification

        A BacktestSpecification includes only the attributes users can directly control, not those
        indirectly determined by the project dataset.

        Returns
        -------
        BacktestSpecification
            the specification for this backtest
        """
        return BacktestSpecification(self.index, self.gap_duration,
                                     self.validation_start_date, self.validation_duration)

    def to_dataframe(self):
        """ Render this backtest as a dataframe for convenience of display

        Returns
        -------
        backtest_partitioning : pandas.Dataframe
            the backtest attributes, formatted into a dataframe
        """
        display_dict = {
            'start_date': {
                'backtest_{}_available_training'.format(self.index):
                    self.available_training_start_date,
                'backtest_{}_primary_training'.format(self.index):
                    self.primary_training_start_date,
                'backtest_{}_gap'.format(self.index): self.gap_start_date,
                'backtest_{}_validation'.format(self.index): self.validation_start_date
            },
            'duration': {
                'backtest_{}_available_training'.format(self.index):
                    self.available_training_duration,
                'backtest_{}_primary_training'.format(self.index): self.primary_training_duration,
                'backtest_{}_gap'.format(self.index): self.gap_duration,
                'backtest_{}_validation'.format(self.index): self.validation_duration
            },
            'end_date': {
                'backtest_{}_available_training'.format(self.index):
                    self.available_training_end_date,
                'backtest_{}_primary_training'.format(self.index): self.primary_training_end_date,
                'backtest_{}_gap'.format(self.index): self.gap_end_date,
                'backtest_{}_validation'.format(self.index): self.validation_end_date
            }
        }
        return pd.DataFrame.from_dict(display_dict)

    @classmethod
    def from_data(cls, data):
        data = _backtest_converter.check(data)
        return cls(**data)


class DatetimePartitioningSpecification(PartitioningMethod):
    """ Uniquely defines a DatetimePartitioning for some project

    Includes only the attributes of DatetimePartitioning that are directly controllable by users,
    not those determined by the DataRobot application based on the project dataset and the
    user-controlled settings.

    This is the specification that should be passed to `Project.set_target` via the
    `partitioning_method` parameter.  To see the full partitioning based on the project dataset,
    use `DatetimePartitioning.generate`.

    All durations should be specified with a duration string such as those returned
    by the `partitioning_methods.construct_duration_string` helper method.

    Attributes
    ----------
    datetime_partition_column : str
        the name of the column whose values as dates are used to assign a row
        to a particular partition
    autopilot_data_selection_method : str
        one of ``datarobot.enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD``.  Whether models created
        by the autopilot should use "rowCount" or "duration" as their data_selection_method.
    validation_duration : str or None
        the default validation_duration for the backtests
    holdout_start_date : datetime.datetime or None
        The start date of holdout scoring data.  If holdout_start_date is specified,
        holdout_duration must also be specified.
    holdout_duration : str or None
        The duration of the holdout scoring data.  If holdout_duration is specified,
        holdout_start_date must also be specified.
    gap_duration : str or None
        The duration of the gap between training and holdout scoring data
    number_of_backtests : int or None
        the number of backtests to  use
    backtests : list of BacktestSpecification
        the exact specification of backtests to use.  The indexes of the specified backtests should
        range from 0 to number_of_backtests - 1.  If any backtest is left unspecified, a default
        configuration will be chosen.
    """
    def __init__(self, datetime_partition_column, autopilot_data_selection_method=None,
                 validation_duration=None,
                 holdout_start_date=None, holdout_duration=None,
                 gap_duration=None, number_of_backtests=None, backtests=None):
        self.datetime_partition_column = datetime_partition_column
        self.autopilot_data_selection_method = autopilot_data_selection_method
        self.validation_duration = validation_duration
        self.holdout_start_date = holdout_start_date
        self.holdout_duration = holdout_duration
        self.gap_duration = gap_duration
        self.number_of_backtests = number_of_backtests
        self.backtests = backtests or []

    def collect_payload(self):
        if self.holdout_start_date and not isinstance(self.holdout_start_date, datetime):
            raise ValueError('expected holdout_start_date to be a datetime.datetime')
        return {
            'datetime_partition_column': self.datetime_partition_column,
            'autopilot_data_selection_method': self.autopilot_data_selection_method,
            'validation_duration': self.validation_duration,
            'holdout_start_date': self.holdout_start_date,
            'holdout_duration': self.holdout_duration,
            'gap_duration': self.gap_duration,
            'number_of_backtests': self.number_of_backtests,
            'backtests': [bt.collect_payload() for bt in self.backtests] or None,
            'cv_method': CV_METHOD.DATETIME
        }


class DatetimePartitioning(object):
    """ Full partitioning of a project for datetime partitioning

    Includes both the attributes specified by the user, as well as those determined by the DataRobot
    application based on the project dataset.  In order to use a partitioning to set the target,
    call `to_specification` and pass the resulting `DatetimePartitioningSpecification` to
    `Project.set_target`.

    The available training data corresponds to all the data available for training, while the
    primary training data corresponds to the data that can be used to train while ensuring that all
    backtests are available.  If a model is trained with more data than is available in the primary
    training data, then all backtests may not have scores available.

    Attributes
    ----------
    project_id : str
        the id of the project this partitioning applies to
    datetime_partition_column : str
        the name of the column whose values as dates are used to assign a row
        to a particular partition
    date_format : str
        the format (e.g. "%Y-%m-%d %H:%M:%S") by which the partition column was interpreted
        (compatible with strftime [https://docs.python.org/2/library/time.html#time.strftime] )
    autopilot_data_selection_method : str
        one of ``datarobot.enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD``.  Whether models created
        by the autopilot use "rowCount" or "duration" as their data_selection_method.
    validation_duration : str
        the validation duration specified when initializing the partitioning - not directly
        significant if the backtests have been modified, but used as the default validation_duration
        for the backtests
    available_training_start_date : datetime.datetime
        The start date of the available training data for scoring the holdout
    available_training_duration : str
        The duration of the available training data for scoring the holdout
    available_training_row_count : int or None
        The number of rows in the available training data for scoring the holdout.  Only available
        when retrieving the partitioning after setting the target.
    available_training_end_date : datetime.datetime
        The end date of the available training data for scoring the holdout
    primary_training_start_date : datetime.datetime
        The start date of primary training data for scoring the holdout
    primary_training_duration : str
        The duration of the primary training data for scoring the holdout
    primary_training_row_count : int or None
        The number of rows in the primary training data for scoring the holdout.  Only available
        when retrieving the partitioning after setting the target.
    primary_training_end_date : datetime.datetime
        The end date of the primary training data for scoring the holdout
    gap_start_date : datetime.datetime
        The start date of the gap between training and holdout scoring data
    gap_duration : str
        The duration of the gap between training and holdout scoring data
    gap_row_count : int or None
        The number of rows in the gap between training and holdout scoring data.  Only available
        when retrieving the partitioning after setting the target.
    gap_end_date : datetime.datetime
        The end date of the gap between training and holdout scoring data
    holdout_start_date : datetime.datetime
        The start date of holdout scoring data
    holdout_duration : str
        The duration of the holdout scoring data
    holdout_row_count : int or None
        The number of rows in the holdout scoring data.  Only available when retrieving the
        partitioning after setting the target.
    holdout_end_date : datetime.datetime
        The end date of the holdout scoring data
    number_of_backtests : int
        the number of backtests used
    backtests : list of partitioning_methods.Backtest
        the configured Backtests
    total_row_count : int
        the number of rows in the project dataset.  Only available when retrieving the partitioning
        after setting the target.
    """
    _client = staticproperty(get_client)
    _converter = t.Dict({
        t.Key('project_id'): t.String(),
        t.Key('datetime_partition_column'): t.String(),
        t.Key('date_format'): t.String(),
        t.Key('autopilot_data_selection_method'): t.String(),
        t.Key('validation_duration'): t.String(),
        t.Key('available_training_start_date'): parse_time,
        t.Key('available_training_duration'): t.String(),
        t.Key('available_training_row_count', optional=True): t.Int(),
        t.Key('available_training_end_date'): parse_time,
        t.Key('primary_training_start_date'): parse_time,
        t.Key('primary_training_duration'): t.String(),
        t.Key('primary_training_row_count', optional=True): t.Int(),
        t.Key('primary_training_end_date'): parse_time,
        t.Key('gap_start_date'): parse_time,
        t.Key('gap_duration'): t.String(),
        t.Key('gap_row_count', optional=True): t.Int(),
        t.Key('gap_end_date'): parse_time,
        t.Key('holdout_start_date'): parse_time,
        t.Key('holdout_duration'): t.String(),
        t.Key('holdout_row_count', optional=True): t.Int(),
        t.Key('holdout_end_date'): parse_time,
        t.Key('number_of_backtests'): t.Int(),
        t.Key('backtests'): t.List(_backtest_converter),
        t.Key('total_row_count', optional=True): t.Int()
    }).ignore_extra('*')

    def __init__(self, project_id=None, datetime_partition_column=None, date_format=None,
                 autopilot_data_selection_method=None,
                 validation_duration=None, available_training_start_date=None,
                 available_training_duration=None, available_training_row_count=None,
                 available_training_end_date=None,
                 primary_training_start_date=None, primary_training_duration=None,
                 primary_training_row_count=None, primary_training_end_date=None,
                 gap_start_date=None, gap_duration=None, gap_row_count=None, gap_end_date=None,
                 holdout_start_date=None, holdout_duration=None, holdout_row_count=None,
                 holdout_end_date=None,
                 number_of_backtests=None, backtests=None, total_row_count=None):
        self.project_id = project_id
        self.datetime_partition_column = datetime_partition_column
        self.date_format = date_format
        self.autopilot_data_selection_method = autopilot_data_selection_method
        self.validation_duration = validation_duration
        self.available_training_start_date = available_training_start_date
        self.available_training_duration = available_training_duration
        self.available_training_row_count = available_training_row_count
        self.available_training_end_date = available_training_end_date
        self.primary_training_start_date = primary_training_start_date
        self.primary_training_duration = primary_training_duration
        self.primary_training_row_count = primary_training_row_count
        self.primary_training_end_date = primary_training_end_date
        self.gap_start_date = gap_start_date
        self.gap_duration = gap_duration
        self.gap_row_count = gap_row_count
        self.gap_end_date = gap_end_date
        self.holdout_start_date = holdout_start_date
        self.holdout_duration = holdout_duration
        self.holdout_row_count = holdout_row_count
        self.holdout_end_date = holdout_end_date
        self.number_of_backtests = number_of_backtests
        self.backtests = backtests
        self.total_row_count = total_row_count

    @classmethod
    def from_server_data(cls, data):
        converted_data = cls._converter.check(from_api(data))
        converted_data['backtests'] = [Backtest(**backtest_data) for backtest_data
                                       in converted_data['backtests']]
        return cls(**converted_data)

    @classmethod
    def generate(cls, project_id, spec):
        """ Preview the full partitioning determined by a DatetimePartitioningSpecification

        Based on the project dataset and the partitioning specification, inspect the full
        partitioning that would be used if the same specification were passed into
        `Project.set_target`.

        Parameters
        ----------
        project_id : str
            the id of the project
        spec : DatetimePartitioningSpec
            the desired partitioning

        Returns
        -------
        DatetimePartitioning :
            the full generated partitioning
        """
        url = 'projects/{}/datetimePartitioning/'.format(project_id)
        payload = spec.collect_payload()
        payload.pop('cv_method')
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def get(cls, project_id):
        """ Retrieve the DatetimePartitioning from a project

        Only available if the project has already set the target as a datetime project.

        Parameters
        ----------
        project_id : str
            the id of the project to retrieve partitioning for

        Returns
        -------
        DatetimePartitioning : the full partitioning for the project
        """
        url = 'projects/{}/datetimePartitioning/'.format(project_id)
        response = cls._client.get(url)
        return cls.from_server_data(response.json())

    def to_specification(self):
        """ Render the DatetimePartitioning as a DatetimePartitioningSpecification

        The resulting specification can be used when setting the target, and contains only the
        attributes directly controllable by users.

        Returns
        -------
        DatetimePartitioningSpecification:
            the specification for this partitioning
        """
        return DatetimePartitioningSpecification(
            self.datetime_partition_column,
            autopilot_data_selection_method=self.autopilot_data_selection_method,
            validation_duration=self.validation_duration,
            holdout_start_date=self.holdout_start_date, holdout_duration=self.holdout_duration,
            gap_duration=self.gap_duration, number_of_backtests=self.number_of_backtests,
            backtests=[bt.to_specification() for bt in self.backtests]
        )

    def to_dataframe(self):
        """ Render the partitioning settings as a dataframe for convenience of display

        Excludes project_id, datetime_partition_column, date_format,
        autopilot_data_selection_method, validation_duration,
        and number_of_backtests, as well as the row count information, if present.
        """
        display_dict = {'start_date': {'available_training': self.available_training_start_date,
                                       'primary_training': self.primary_training_start_date,
                                       'gap': self.gap_start_date,
                                       'holdout': self.holdout_start_date},
                        'duration': {'available_training': self.available_training_duration,
                                     'primary_training': self.primary_training_duration,
                                     'gap': self.gap_duration,
                                     'holdout': self.holdout_duration},
                        'end_date': {'available_training': self.available_training_end_date,
                                     'primary_training': self.primary_training_end_date,
                                     'gap': self.gap_end_date,
                                     'holdout': self.holdout_end_date}}
        display_df = pd.DataFrame.from_dict(display_dict)
        final_df = display_df.append([bt.to_dataframe() for bt in self.backtests])
        return final_df



from datetime import datetime

import pandas as pd
import trafaret as t

from ..utils import encode_utf8_if_py2, get_id_from_response
from .api_object import APIObject

int_float_string = (t.Type(int) | t.Type(float) | t.String(allow_blank=True))

prediction_values_trafaret = t.Dict({
    t.Key('label'): int_float_string,
    t.Key('value'): t.Float
}).ignore_extra('*')

reason_code_entry_trafaret = t.Dict({
    t.Key('label'): int_float_string,
    t.Key('feature'): t.String,
    t.Key('feature_value'): int_float_string,
    t.Key('strength'): t.Float,
    t.Key('qualitative_strength'): t.String
}).ignore_extra('*')

reason_code_trafaret = t.Dict({
    t.Key('row_id'): t.Int,
    t.Key('prediction'): int_float_string,
    t.Key('prediction_values'): t.List(prediction_values_trafaret),
    t.Key('reason_codes'): t.List(reason_code_entry_trafaret)
}).ignore_extra('*')


class ReasonCodesInitialization(APIObject):
    """
    Represents a reason codes initialization of a model.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model reason codes initialization is for
    reason_codes_sample : list of dict
        a small sample of reason codes that could be generated for the model
    """
    _path_template = 'projects/{}/models/{}/reasonCodesInitialization/'
    _converter = t.Dict({
        t.Key('project_id'): t.String,
        t.Key('model_id'): t.String,
        t.Key('reason_codes_sample'): t.List(reason_code_trafaret),
    }).allow_extra('*')

    def __init__(self, project_id, model_id, reason_codes_sample=None):
        self.project_id = project_id
        self.model_id = model_id
        self.reason_codes_sample = reason_codes_sample

        self._path = self._path_template.format(self.project_id, self.model_id)

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(project_id={}, model_id={})'.format(type(self).__name__,
                                                                           self.project_id,
                                                                           self.model_id))

    @classmethod
    def get(cls, project_id, model_id):
        """
        Retrieve the reason codes initialization for a model.

        Reason codes initializations are a prerequisite for computing reason codes, and include
        a sample what the computed reason codes for a prediction dataset would look like.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model reason codes initialization is for

        Returns
        -------
        reason_codes_initialization : ReasonCodesInitialization
            The queried instance.
        """
        path = cls._path_template.format(project_id, model_id)
        return cls.from_location(path)

    @classmethod
    def create(cls, project_id, model_id):
        """
        Create a reason codes initialization for the specified model.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which initialization is requested

        Returns
        -------
        job : Job
            an instance of created async job
        """
        from .job import Job
        response = cls._client.post(cls._path_template.format(project_id, model_id))
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    def delete(self):
        """
        Delete this reason codes initialization.
        """
        self._client.delete(self._path)


class ReasonCodes(APIObject):
    """
    Represents reason codes metadata and provides access to computation results.

    Examples
    --------
    .. code-block:: python

        reason_codes = dr.ReasonCodes.get(project_id, reason_codes_id)
        for row in reason_codes.get_rows():
            print(row)  # row is an instance of ReasonCodesRow

    Attributes
    ----------
    id : str
        id of the record and reason codes computation result
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model reason codes initialization is for
    dataset_id : str
        id of the prediction dataset reason codes were computed for
    max_codes : int
        maximum number of reason codes to supply per row of the dataset
    threshold_low : float
        the lower threshold, below which a prediction must score in order for reason codes to be
        computed for a row in the dataset
    threshold_high : fload
        the high threshold, above which a prediction must score in order for reason codes to be
        computed for a row in the dataset
    num_columns : int
        the number of columns reason codes were computed for
    finish_time : float
        timestamp referencing when computation for these reason codes finished
    reason_codes_location : str
        where to retrieve the reason codes
    """
    _path_template = 'projects/{}/reasonCodesRecords/'
    _codes_path_template = 'projects/{}/reasonCodes/'
    _converter = t.Dict({
        t.Key('id'): t.String,
        t.Key('project_id'): t.String,
        t.Key('model_id'): t.String,
        t.Key('dataset_id'): t.String,
        t.Key('max_codes'): t.Int,
        t.Key('threshold_low', optional=True): t.Float,
        t.Key('threshold_high', optional=True): t.Float,
        t.Key('num_columns'): t.Int,
        t.Key('finish_time'): t.Float,
        t.Key('reason_codes_location'): t.String,
    }).allow_extra('*')

    def __init__(self, id, project_id, model_id, dataset_id, max_codes, num_columns, finish_time,
                 reason_codes_location, threshold_low=None, threshold_high=None):
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.max_codes = max_codes
        self.threshold_low = threshold_low
        self.threshold_high = threshold_high
        self.id = id
        self.num_columns = num_columns
        self.finish_time = datetime.fromtimestamp(finish_time)
        self.reason_codes_location = reason_codes_location

        self._path = self._path_template.format(self.project_id)

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(id={}, project_id={}, model_id={})'.format(
            type(self).__name__, self.id, self.project_id, self.model_id))

    @classmethod
    def get(cls, project_id, reason_codes_id):
        """
        Retrieve a specific reason codes.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        reason_codes_id : str
            id of the reason codes

        Returns
        -------
        reason_codes : ReasonCodes
            The queried instance.
        """
        path = '{}{}/'.format(cls._path_template.format(project_id), reason_codes_id)
        return cls.from_location(path)

    @classmethod
    def create(cls, project_id, model_id, dataset_id,
               max_codes=None, threshold_low=None, threshold_high=None):
        """
        Create a reason codes for the specified dataset.

        In order to create ReasonCodesPage for a particular model and dataset, you must first:

          * Compute feature impact for the model via ``datarobot.Model.get_feature_impact()``
          * Compute a ReasonCodesInitialization for the model via
            ``datarobot.ReasonCodesInitialization.create(project_id, model_id)``
          * Compute predictions for the model and dataset via
            ``datarobot.Model.request_predictions(dataset_id)``

        ``threshold_high`` and ``threshold_low`` are optional filters applied to speed up
        computation.  When at least one is specified, only the selected outlier rows will have
        reason codes computed. Rows are considered to be outliers if their predicted
        value (in case of regression projects) or probability of being the positive
        class (in case of classification projects) is less than ``threshold_low`` or greater than
        ``thresholdHigh``.  If neither is specified, reason codes will be computed for all rows.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which reason codes are requested
        dataset_id : str
            id of the prediction dataset for which reason codes are requested
        threshold_low : float, optional
            the lower threshold, below which a prediction must score in order for reason codes to be
            computed for a row in the dataset. If neither ``threshold_high`` nor ``threshold_low``
            is specified, reason codes will be computed for all rows.
        threshold_high : float, optional
            the high threshold, above which a prediction must score in order for reason codes to be
            computed. If neither ``threshold_high`` nor ``threshold_low`` is specified, reason codes
            will be computed for all rows.
        max_codes : int, optional
            the maximum number of reason codes to supply per row of the dataset, default: 3.

        Returns
        -------
        job: Job
            an instance of created async job
        """
        from .job import Job
        payload = {
            'model_id': model_id,
            'dataset_id': dataset_id,
        }
        if max_codes is not None:
            payload['max_codes'] = max_codes
        if threshold_low is not None:
            payload['threshold_low'] = threshold_low
        if threshold_high is not None:
            payload['threshold_high'] = threshold_high
        response = cls._client.post(cls._codes_path_template.format(project_id), data=payload)
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    @classmethod
    def list(cls, project_id, model_id=None, limit=None, offset=None):
        """
        List of reason codes for a specified project.

        Parameters
        ----------
        project_id : str
            id of the project to list reason codes for
        model_id : str, optional
            if specified, only reason codes computed for this model will be returned
        limit : int or None
            at most this many results are returned, default: no limit
        offset : int or None
            this many results will be skipped, default: 0

        Returns
        -------
        reason_codes : list[ReasonCodes]
        """
        response = cls._client.get(cls._path_template.format(project_id),
                                   params={'model_id': model_id,
                                           'limit': limit,
                                           'offset': offset})
        r_data = response.json()
        return [cls.from_server_data(item) for item in r_data['data']]

    def get_rows(self, batch_size=None):
        """
        Retrieve reason codes rows.

        Parameters
        ----------
        batch_size : int
            maximum number of reason codes rows to retrieve per request

        Yields
        ------
        reason_codes_row : ReasonCodesRow
            Represents reason codes computed for a prediction row.
        """
        page = self.get_reason_codes_page(limit=batch_size)
        while True:
            for row in page.data:
                yield ReasonCodesRow(**row)
            if not page.next_page:
                break
            page = ReasonCodesPage.from_location(page.next_page)

    def get_all_as_dataframe(self):
        """
        Retrieve all reason codes rows and return them as a pandas.DataFrame.

        Returned dataframe has the following structure:

            - row_id : row id from prediction dataset
            - prediction : the output of the model for this row
            - class_0_label : a class level from the target (only appears for classification
              projects)
            - class_0_probability : the probability that the target is this class (only appears for
              classification projects)
            - class_1_label : a class level from the target (only appears for classification
              projects)
            - class_1_probability : the probability that the target is this class (only appears for
              classification projects)
            - reason_0_feature : the name of the feature contributing to the prediction for this
              reason
            - reason_0_feature_value : the value the feature took on
            - reason_0_label : the output being driven by this reason.  For regression projects,
              this is the name of the target feature.  For classification projects, this is the
              class label whose probability increasing would correspond to a positive strength.
            - reason_0_qualitative_strength : a human-readable description of how strongly the
              feature affected the prediction (e.g. '+++', '--', '+') for this reason
            - reason_0_strength : the amount this feature's value affected the prediction
            - ...
            - reason_N_feature : the name of the feature contributing to the prediction for this
              reason
            - reason_N_feature_value : the value the feature took on
            - reason_N_label : the output being driven by this reason.  For regression projects,
              this is the name of the target feature.  For classification projects, this is the
              class label whose probability increasing would correspond to a positive strength.
            - reason_N_qualitative_strength : a human-readable description of how strongly the
              feature affected the prediction (e.g. '+++', '--', '+') for this reason
            - reason_N_strength : the amount this feature's value affected the prediction

        Returns
        -------
        dataframe: pandas.DataFrame
        """
        columns = ['row_id', 'prediction']
        rows = self.get_rows(batch_size=1)
        first_row = next(rows)
        # for regression, length is 1; for classification, length is number of levels in target
        # i.e. 2 for binary classification
        is_classification = len(first_row.prediction_values) > 1
        # include class label/probability for classification project
        if is_classification:
            for i in range(len(first_row.prediction_values)):
                columns.extend(['class_{}_label'.format(i), 'class_{}_probability'.format(i)])
        for i in range(self.max_codes):
            columns.extend(['reason_{}_feature'.format(i),
                            'reason_{}_feature_value'.format(i),
                            'reason_{}_label'.format(i),
                            'reason_{}_qualitative_strength'.format(i),
                            'reason_{}_strength'.format(i)])
        rc_list = []

        for i, row in enumerate(self.get_rows()):
            data = [row.row_id, row.prediction]
            if is_classification:
                for pred_value in row.prediction_values:
                    data.extend([pred_value['label'], pred_value['value']])
            for rc in row.reason_codes:
                data.extend([rc['feature'],
                             rc['feature_value'],
                             rc['label'],
                             rc['qualitative_strength'],
                             rc['strength']])
            rc_list.append(data + [None] * (len(columns) - len(data)))

        return pd.DataFrame(data=rc_list, columns=columns)

    def download_to_csv(self, filename, encoding='utf-8'):
        """
        Save reason codes rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save reason codes rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to 'utf-8'
        """
        df = self.get_all_as_dataframe()
        df.to_csv(path_or_buf=filename, header=True, index=False, encoding=encoding)

    def get_reason_codes_page(self, limit=None, offset=None):
        """
        Get reason codes.

        If you don't want use a generator interface, you can access paginated reason codes directly.

        Parameters
        ----------
        limit : int or None
            the number of records to return, the server will use a (possibly finite) default if not
            specified
        offset : int or None
            the number of records to skip, default 0

        Returns
        -------
        reason_codes : ReasonCodesPage
        """
        kwargs = {'limit': limit}
        if offset:
            kwargs['offset'] = offset
        return ReasonCodesPage.get(self.project_id, self.id, **kwargs)

    def delete(self):
        """
        Delete this reason codes.
        """
        path = '{}{}/'.format(self._path_template.format(self.project_id), self.id)
        self._client.delete(path)


class ReasonCodesRow(object):
    """
    Represents reason codes computed for a prediction row.

    Notes
    -----

    ``PredictionValue`` contains:

    * ``label`` : describes what this model output corresponds to.  For regression projects,
      it is the name of the target feature.  For classification projects, it is a level from
      the target feature.
    * ``value`` : the output of the prediction.  For regression projects, it is the predicted
      value of the target.  For classification projects, it is the predicted probability the
      row belongs to the class identified by the label.


    ``ReasonCode`` contains:

    * ``label`` : described what output was driven by this reason code.  For regression
      projects, it is the name of the target feature.  For classification projects, it is the
      class whose probability increasing would correspond to a positive strength of this
      reason code.
    * ``feature`` : the name of the feature contributing to the prediction
    * ``feature_value`` : the value the feature took on for this row
    * ``strength`` : the amount this feature's value affected the prediction
    * ``qualitativate_strength`` : a human-readable description of how strongly the feature
      affected the prediction (e.g. '+++', '--', '+')

    Attributes
    ----------
    row_id : int
        which row this ``ReasonCodeRow`` describes
    prediction : float
        the output of the model for this row
    prediction_values : list
        an array of dictionaries with a schema described as ``PredictionValue``
    reason_codes : list
        an array of dictionaries with a schema described as ``ReasonCode``
    """
    def __init__(self, row_id, prediction, prediction_values, reason_codes=None):
        self.row_id = row_id
        self.prediction = prediction
        self.prediction_values = prediction_values
        self.reason_codes = reason_codes

    def __repr__(self):
        return '{}(row_id={}, prediction={})'.format(type(self).__name__,
                                                     self.row_id,
                                                     self.prediction)


class ReasonCodesPage(APIObject):
    """
    Represents batch of reason codes received by one request.

    Attributes
    ----------
    id : str
        id of the reason codes computation result
    data : list[dict]
        list of raw reason codes, each row corresponds to a row of the prediction dataset
    count : int
        total number of rows computed
    previous_page : str
        where to retrieve previous page of reason codes, None if current page is the first
    next_page : str
        where to retrieve next page of reason codes, None if current page is the last
    reason_codes_record_location : str
        where to retrieve the reason codes metadata
    """
    _path_template = 'projects/{}/reasonCodes/'
    _converter = t.Dict({
        t.Key('id'): t.String,
        t.Key('count'): t.Int,
        t.Key('previous', optional=True): t.String(),
        t.Key('next', optional=True): t.String(),
        t.Key('data'): t.List(reason_code_trafaret),
        t.Key('reason_codes_record_location'): t.URL
    }).allow_extra('*')

    def __init__(self, id, count=None, previous=None, next=None, data=None,
                 reason_codes_record_location=None):
        self.id = id
        self.count = count
        self.previous_page = previous
        self.next_page = next
        self.data = data
        self.reason_codes_record_location = reason_codes_record_location

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(id={})'.format(type(self).__name__, self.id))

    @classmethod
    def get(cls, project_id, reason_codes_id, limit=None, offset=0):
        """
        Retrieve reason codes.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        reason_codes_id : str
            id of the reason codes
        limit : int or None
            the number of records to return, the server will use a (possibly finite) default if not
            specified
        offset : int or None
            the number of records to skip, default 0

        Returns
        -------
        reason_codes : ReasonCodesPage
            The queried instance.
        """
        params = {'offset': offset}
        if limit:
            params['limit'] = limit
        path = '{}{}/'.format(cls._path_template.format(project_id), reason_codes_id)
        return cls.from_location(path, params=params)

    @classmethod
    def from_location(cls, path, params=None):
        server_data = cls._client.get(path, params=params).json()
        return cls.from_server_data(server_data)



import mock
import os

import pytest
import six
import responses
import tempfile
import unittest

import datarobot as dr
from datarobot.rest import RESTClientObject
from datarobot.client import set_client, get_client, Client
from datarobot.errors import AppPlatformError
from .utils import SDKTestcase

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


class ClientTest(object):

    @pytest.fixture(autouse=True)
    def empty_client(self):
        set_client(None)

    @pytest.yield_fixture(autouse=True)
    def no_version_check(self):
        with mock.patch('datarobot.client._version_warning'):
            yield

    def test_instantiation(self):
        """
        Basic client installation.
        """
        client = Client(token='t-token', endpoint='https://host_name.com')
        assert get_client() is client

    @mock.patch('datarobot.client._file_exists', return_value=False)
    def test_instantiation_without_env(self, mock_file_args):
        """
        Basic client installation by get_client without configuration.
        """
        with mock.patch(
            'os.environ',
                {'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            with pytest.raises(ValueError):
                get_client()

    @mock.patch('datarobot.client._file_exists')
    def test_no_endpoint_fails(self, mock_file_exists):
        mock_file_exists.return_value = False
        with mock.patch('os.environ', {}):
            with pytest.raises(ValueError):
                Client(token='NOTATOKEN')

    def test_token_and_endpoint_is_okay(self):
        Client(token='token', endpoint='https://need_an_endpoint.com')

    def test_re_instantiation(self):
        """
        Client re installation.
        """
        previous = Client(token='t-****', endpoint='https://end.com')
        old_client = set_client(RESTClientObject(auth=('u-**********', 'p-******'),
                                                 endpoint='https://host_name.com'))
        assert previous is old_client

    @responses.activate
    def test_recognizing_domain_on_instance(self):
        raw = """{"api_token": "some_token"}"""
        responses.add(responses.POST,
                      'https://host_name.com/api/v2/api_token/',
                      body=raw,
                      status=201,
                      content_type='application/json')
        set_client(RESTClientObject(auth=('u-**********', 'p-******'),
                                    endpoint='https://host_name.com/api/v2'))
        restored_client = get_client()
        assert restored_client.domain == 'https://host_name.com'

    @mock.patch('datarobot.client._file_exists', return_value=False)
    def test_instantiation_from_env(self, mock_file_exists):
        """
        Test instantiation with creds from virtual environment
        """
        with patch.dict(
            'os.environ',
            {'DATAROBOT_API_TOKEN': 'venv_token',
             'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            rest_client = get_client()
            assert rest_client.token == 'venv_token'
            assert rest_client.endpoint == 'https://host_name.com'

        set_client(None)

        with patch.dict(
            'os.environ',
            {'DATAROBOT_API_TOKEN': 'venv_token',
             'DATAROBOT_ENDPOINT': 'https://host_name.com'}):
            rest_client = get_client()
            assert rest_client.token == 'venv_token'

    def test_instantiation_from_file_with_wrong_path(self):
        with patch.dict('os.environ', {'DATAROBOT_CONFIG_FILE': './tests/fixtures/.datarobotrc'}):
            with pytest.raises(ValueError):
                get_client()

    def test_instantiation_from_yaml_file_api_token(self):
        file_data = ('token: fake_token\n'
                     'endpoint: https://host_name.com')
        config_file = tempfile.NamedTemporaryFile(delete=False, suffix='.yaml')
        try:
            with open(config_file.name, mode='w') as config:
                config.write(file_data)
            fake_environ = {'DATAROBOT_CONFIG_FILE': config_file.name}
            with patch('os.environ', fake_environ):
                rest_client = get_client()
            assert rest_client.token == 'fake_token'
            assert rest_client.endpoint == 'https://host_name.com'
        finally:
            os.remove(config_file.name)

    def test_client_from_codeline(self):
        Client(token='some_token',
               endpoint='https://endpoint.com')
        c = get_client()
        assert c.token == 'some_token'
        assert c.endpoint == 'https://endpoint.com'


class RestErrors(SDKTestcase):

    @responses.activate
    def test_404_plain_text(self):
        """
        Bad request in plain text
        """
        raw = "Not Found"

        responses.add(responses.GET, 'https://host_name.com/projects/404',
                      body=raw, status=404, content_type='text/plain')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/404')

        self.assertEqual(str(app_error.exception),
                         '404 client error: Not Found')

    @responses.activate
    def test_404_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(responses.GET, 'https://host_name.com/projects/404',
                      body=raw, status=404, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/404')

        exception_message = str(app_error.exception)
        self.assertIn('404 client error', exception_message)
        self.assertIn('Not Found', exception_message)

    @responses.activate
    def test_500_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(responses.GET, 'https://host_name.com/projects/500',
                      body=raw, status=500, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/500')

        exception_message = str(app_error.exception)
        self.assertIn('500 server error', exception_message)
        self.assertIn('Not Found', exception_message)

    @responses.activate
    def test_other_errors(self):
        """
        Other errors
        """
        raw = """
        {"message": "Bad response"}
        """

        responses.add(responses.GET,
                      'https://host_name.com/projects/500',
                      body=raw, status=500, content_type='application/json')

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get('projects/500')
            self.assertEqual(str(app_error.exception),
                             'Connection refused: '
                             'https://host_name.com/projects/500')


class TestClientAttributes(unittest.TestCase):

    def test_main_useful_things_under_datarobot(self):
        """To lower the intimidation factor, let's try to limit the objects
        that show up at the root of datarobot

        This way, when they are in IPython and do tab-completion they get a
        sense for what is available to tinker with at the top level
        """
        known_names = {'Project',
                       'Model',
                       'DatetimeModel',
                       'PrimeModel',
                       'BlenderModel',
                       'FrozenModel',
                       'Ruleset',
                       'Blueprint',
                       'ModelJob',
                       'PredictJob',
                       'Job',
                       'PredictionDataset',
                       'PrimeFile',
                       'QUEUE_STATUS',
                       'Client',
                       'AUTOPILOT_MODE',
                       'AppPlatformError',
                       'utils',
                       'errors',
                       'models',
                       'client',
                       'rest',
                       'async',
                       'SCORING_TYPE',
                       'Featurelist',
                       'Feature',
                       'helpers',
                       'RandomCV',
                       'StratifiedCV',
                       'GroupCV',
                       'UserCV',
                       'RandomTVH',
                       'UserTVH',
                       'StratifiedTVH',
                       'GroupTVH',
                       'DatetimePartitioning',
                       'DatetimePartitioningSpecification',
                       'BacktestSpecification',
                       'partitioning_methods',
                       'AdvancedOptions',
                       'VERBOSITY_LEVEL',
                       'enums',
                       'ImportedModel',
                       'ReasonCodesInitialization',
                       'ReasonCodes',
                       }

        found_names = {name for name in dir(dr)
                       if not name.startswith('_')}
        assert found_names == known_names


@pytest.yield_fixture
def mock_os():
    with patch('datarobot.client.os', autospec=True) as m:
        yield m


@pytest.yield_fixture
def mock_file_exists():
    with patch('datarobot.client._file_exists') as m:
        yield m


@pytest.yield_fixture
def mock_os_environ():
    with patch('datarobot.client.os.environ') as m:
        yield m



import trafaret as t

from datarobot.models.api_object import APIObject


class Dummy(APIObject):

    _converter = t.Dict({
        t.Key('renamed') >> 'magic': t.String()
    })

    def __init__(self, magic=None):
        self.magic = magic


def test_safe_data():
    data = Dummy._safe_data({'renamed': 'nombre'})
    dummy = Dummy(**data)
    assert dummy.magic == 'nombre'



import six
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import deprecation_warning
from ..utils import encode_utf8_if_py2


class Featurelist(APIObject):
    """ A set of features used in modeling

    Attributes
    ----------
    id : str
        the id of the featurelist
    name : str
        the name of the featurelist
    features : list of str
        the names of all the Features in the Featurelist
    project_id : str
        the project the Featurelist belongs to
    """
    _path = 'projects/{}/featurelists/'
    _converter = t.Dict({
            t.Key('id', optional=True) >> 'id': t.String(allow_blank=True),
            t.Key('name', optional=True) >> 'name': t.String,
            t.Key('features', optional=True) >> 'features': t.List(t.String),
            t.Key('project_id', optional=True): t.String()
        }).allow_extra("*")

    def __repr__(self):
        return encode_utf8_if_py2(u'Featurelist({})'.format(self.name))

    def __init__(self, id=None, name=None, features=None, project_id=None):
        if isinstance(id, dict):
            deprecation_warning('Featurelist instantiation from a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                )
            self.__init__(**id)
        else:
            self.id = id
            self.name = name
            self.features = features
            self._project = None
            self.project_id = project_id
            self._set_up(self.project_id)

    @property
    def project(self):
        deprecation_warning('Featurelist.project',
                            deprecated_since_version='v2.3',
                            will_remove_version='v3.0',
                            message='Users should construct a Project from the project_id')
        return self._project

    def _set_up(self, project_id):
        from . import Project
        self._project = Project(project_id)

    @classmethod
    def get(cls, project_id, featurelist_id):
        """Retrieve a known feature list

        Parameters
        ----------
        project_id : str
            The id of the project the featurelist is associated with
        featurelist_id : str
            The ID of the featurelist to retrieve

        Returns
        -------
        featurelist : Featurelist
            The queried instance
        """
        from . import Project

        if isinstance(project_id, Project):
            deprecation_warning('Featurelist.get using an instance of Project',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use a project_id string instead')
            project_id = project_id.id
        elif isinstance(project_id, six.string_types):
            pass
        else:
            raise ValueError('Project arg must be Project instance or str')

        url = cls._path.format(project_id) + featurelist_id + '/'
        return cls.from_location(url)



import json

import pytest
import responses

from datarobot import errors, Project


@responses.activate
@pytest.mark.usefixtures('client')
def test_timeout(mock_async_time, async_url):

    mock_async_time.time.side_effect = (0, 0, 1000)

    responses.add(
        responses.GET,
        async_url,
        content_type='application/json',
        status=200,
        body=json.dumps({'status': 'RUNNING'}),
    )

    with pytest.raises(errors.AsyncTimeoutError):
        Project.from_async(async_url, max_wait=1)


@responses.activate
@pytest.mark.usefixtures('client')
def test_check_failed(async_url):
    """
    Getting status_code different from 200 and 303 should raise AsyncFailureError
    """
    responses.add(
        responses.GET,
        async_url,
        content_type='application/json',
        status=504,
        body=json.dumps({}),
    )

    with pytest.raises(errors.AsyncFailureError):
        Project.from_async(async_url, max_wait=1)


@responses.activate
@pytest.mark.usefixtures('client')
def test_check_exception_for_failure_contains_status_code_and_async_location(async_url):
    """
    Getting status_code different from 200 and 303 should raise AsyncFailureError
    """
    responses.add(
        responses.GET,
        async_url,
        content_type='application/json',
        status=504,
        body=json.dumps({}),
    )

    with pytest.raises(errors.ProjectAsyncFailureError) as e:
        Project.from_async(async_url, max_wait=1)
        assert e.status_code == 504
        assert e.async_location == async_url


@responses.activate
@pytest.mark.usefixtures('client')
def test_project_creation_failed(async_url,
                                 async_failure_json):
    responses.add(
        responses.GET,
        async_url,
        content_type='application/json',
        status=200,
        body=async_failure_json,
    )

    with pytest.raises(errors.AsyncProcessUnsuccessfulError):
        Project.from_async(async_url, max_wait=1)


@responses.activate
@pytest.mark.usefixtures('client')
def test_project_creation_aborted(async_url,
                                  async_aborted_json):
    responses.add(
        responses.GET,
        async_url,
        content_type='application/json',
        status=200,
        body=async_aborted_json,
    )

    with pytest.raises(errors.AsyncProcessUnsuccessfulError):
        Project.from_async(async_url, max_wait=1)



import json

import pytest
import responses

from datarobot import Feature, Project


@pytest.fixture
def list_features_json():
    return """
    [{
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f"
    }]
    """


@pytest.fixture
def feature_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": 1,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f"
    }
    """


@pytest.fixture
def feature_sans_importance_json():
    return """
    {
        "id": 34,
        "name": "Claim_Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f"
    }
    """


@pytest.fixture
def claim_amount_json():
    return """
    {
        "id": 34,
        "name": "Claim Amount",
        "featureType": "Numeric",
        "importance": null,
        "lowInformation": false,
        "uniqueCount": 200,
        "naCount": 0,
        "dateFormat": null,
        "projectId": "56cdd3baccf94e69733ff20f"
    }
    """


@pytest.fixture
def low_info_feature_json():
    """The information for a feature that has been marked as low info

    DataRobot does not finish gathering the naCount or generate a featureType
    in these situations
    """
    return """
    {
    "featureType": null,
    "lowInformation": true,
    "name": "mths_since_last_major_derog",
    "uniqueCount": 0,
    "importance": null,
    "id": 25,
    "naCount": null,
    "dateFormat": null,
    "projectId": "56cdd3baccf94e69733ff20f"
    }
    """


@pytest.fixture
def date_feature_json():
    return """
    {
        "uniqueCount": 157,
        "lowInformation": false,
        "name": "dates",
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "importance": 0.2648777523946366,
        "projectId": "56cdd3baccf94e69733ff20f",
        "featureType": "Date",
        "id": 1,
        "naCount": 0
    }
    """


@pytest.fixture
def feature_server_data(feature_json):
    return json.loads(feature_json)


def test_future_proof(feature_server_data):
    Feature.from_server_data(dict(feature_server_data, future='key'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_features(list_features_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/',
                  body=list_features_json)
    feature, = Project('p-id').get_features()
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance == 1
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.project_id == '56cdd3baccf94e69733ff20f'


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature(feature_sans_importance_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=feature_sans_importance_json)
    feature = Feature.get('p-id', 'f-id')
    assert isinstance(feature, Feature)
    assert feature.id == 34
    assert feature.name == 'Claim_Amount'
    assert feature.feature_type == 'Numeric'
    assert feature.importance is None
    assert feature.low_information is False
    assert feature.unique_count == 200
    assert feature.na_count == 0
    assert feature.date_format is None
    assert feature.project_id == '56cdd3baccf94e69733ff20f'


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_with_low_info(low_info_feature_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/f-id/',
                  body=low_info_feature_json)
    feature = Feature.get('p-id', 'f-id')
    assert feature.na_count is None
    assert feature.feature_type is None


@responses.activate
@pytest.mark.usefixtures('client')
def test_feature_with_space(claim_amount_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/features/Claim%20Amount/',
                  body=claim_amount_json)
    feature = Feature.get('p-id', 'Claim Amount')
    assert isinstance(feature, Feature)


@responses.activate
@pytest.mark.usefixtures('client')
def test_date_feature(date_feature_json):
    responses.add(responses.GET, 'https://host_name.com/projects/p-id/features/dates/',
                  body=date_feature_json)
    feature = Feature.get('p-id', 'dates')
    assert isinstance(feature, Feature)
    assert feature.date_format == json.loads(date_feature_json)['dateFormat']


def test_feature_with_non_ascii_name(feature_server_data, unicode_string):
    data_copy = dict(feature_server_data)
    data_copy['name'] = unicode_string
    feature = Feature.from_server_data(data_copy)
    print(feature)  # actually part of the test - this used to fail (testing __repr__)


@pytest.mark.usefixtures('known_warning')
def test_bc_instantiate_feature_from_dict(feature_server_data):
    feature = Feature(feature_server_data)
    assert feature.id == feature_server_data['id']
    assert feature.name == feature_server_data['name']
    assert feature.feature_type == feature_server_data['featureType']
    assert feature.importance == feature_server_data['importance']
    assert feature.low_information == feature_server_data['lowInformation']
    assert feature.unique_count == feature_server_data['uniqueCount']
    assert feature.na_count == feature_server_data['naCount']
    assert feature.project_id == '56cdd3baccf94e69733ff20f'



import pytest
import json


@pytest.fixture
def async_url():
    return 'http://host_name.com/status/status_id/'


@pytest.fixture
def resolved_url():
    return 'https://host_name.com/finished/id/'


@pytest.fixture
def async_success_callback(resolved_url):
    def callback(request):
        return 303, {'Location': resolved_url}, ''
    return callback


@pytest.fixture
def async_pending_callback():
    def callback(request):
        return 200, {}, json.dumps(dict(status='INITIALIZED',
                                        message='',
                                        code=0,
                                        created='2016-07-22T12:00:00.123456Z'))
    return callback


@pytest.fixture
def async_job_failure_callback():
    def callback(request):
        return 200, {}, json.dumps(dict(status='ERROR',
                                        message='',
                                        code=0,
                                        created='2016-07-22T12:00:00.123456Z'))
    return callback


@pytest.fixture
def async_server_nonsense_callback():
    """
    A callback that returns a nonsensical status code.
    400-level and 500-level are caught and raised automatically by the client, so
    we'll have this callback return a 299 (that is not a real status code)
    to stand in for the "I don't know what the server is trying to say"
    """
    def callback(request):
        return 299, {}, json.dumps({'message': 'An unexpected error has occurred'})
    return callback


@pytest.fixture
def async_not_finished_callback():
    def callback(request):
        return 200, {}, json.dumps(dict(status='RUNNING',
                                        message='',
                                        code=0,
                                        created='2016-07-22T12:00:00.123456Z'))
    return callback



# coding: utf-8
from datarobot import Model

from ..utils import assert_equal_py2, assert_equal_py3


class TestRepr(object):
    """Tests of method Model.__repr__"""

    def test_model_type_str(self):
        # when both model_type and id are set
        frozen_model = Model(id='id', model_type='model_type')
        # then repr contains only model_type
        assert repr(frozen_model) == "Model('model_type')"

    def test_model_type_unicode(self):
        # when model_type contains non-ascii characters
        frozen_model = Model(model_type=u'тайп')
        # then repr is an ascii string
        assert_equal_py2(repr(frozen_model), "Model(u'\\u0442\\u0430\\u0439\\u043f')")
        assert_equal_py3(repr(frozen_model), "Model('тайп')")

    def test_id(self):
        # when id is set and model_type is missing
        frozen_model = Model(id='58668cb5bf36cd77a5d8056c')
        # then repr contains id
        assert repr(frozen_model) == "Model('58668cb5bf36cd77a5d8056c')"



import json
import responses
import pytest

from datarobot import PredictionDataset, Project
from tests.test_helpers import fixture_file_path


@pytest.fixture
def list_dataset_json():
    return """
    {"count": 1,
     "next": null,
     "previous": null,
     "data": [
         {
             "id": "ds-id",
             "projectId": "p-id",
             "name": "My PredictionDataset",
             "created": "2016-02-16T12:00:00.123456Z",
             "numRows": 100,
             "numColumns": 20
         }
     ]
    }
    """


@pytest.fixture
def dataset_json(list_dataset_json):
    data = json.loads(list_dataset_json)
    return json.dumps(data['data'][0])


dataset_response = {'id': 'ds-id',
                    'projectId': 'p-id',
                    'name': 'My PredictionDataset',
                    'created': '2016-02-16T12:00:00.123456Z',
                    'numRows': 100,
                    'numColumns': 20}

datasets_list_response = {'count': 1,
                          'next': None,
                          'previous': None,
                          'data': [dataset_response]}


def prep_successful_upload_response(upload_type, body):
    responses.add(responses.POST,
                  'https://host_name.com/projects/p-id/predictionDatasets/{}/'.format(upload_type),
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': 'https://host_name.com/status/status-id/'}
                  )
    responses.add(responses.GET,
                  'https://host_name.com/status/status-id/',
                  status=303,
                  body='',
                  content_type='application/json',
                  adding_headers={
                      'Location': 'https://host_name.com/projects/p-id/predictionDatasets/ds-id/'}
                  )
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  status=200,
                  body=body,
                  content_type='application/json')


def assert_matches_expected_dataset(actual_dataset):
    assert actual_dataset.project_id == dataset_response['projectId']
    assert actual_dataset.id == dataset_response['id']
    assert actual_dataset.name == dataset_response['name']
    assert actual_dataset.created == dataset_response['created']
    assert actual_dataset.num_rows == dataset_response['numRows']
    assert actual_dataset.num_columns == dataset_response['numColumns']


def test_future_proof():
    PredictionDataset.from_server_data(dict(dataset_response, new_key='future'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_datasets(list_dataset_json):
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/',
                  body=list_dataset_json)
    project = Project('p-id')
    datasets = project.get_datasets()
    assert len(datasets) == 1
    dataset = datasets[0]
    assert_matches_expected_dataset(dataset)


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve_dataset():
    responses.add(responses.GET,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  body=json.dumps(dataset_response))
    dataset = PredictionDataset.get('p-id', 'ds-id')
    assert_matches_expected_dataset(dataset)


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_url(dataset_json):
    prep_successful_upload_response('urlUploads', dataset_json)
    project = Project('p-id')
    file_url = 'https://my_files.com/my_dataset.csv'
    new_dataset = project.upload_dataset(file_url)
    assert_matches_expected_dataset(new_dataset)
    assert responses.calls[0].request.body == json.dumps({'url': file_url})


@responses.activate
@pytest.mark.usefixtures('client')
def test_upload_dataset_from_file(dataset_json):
    prep_successful_upload_response('fileUploads', dataset_json)
    filepath = fixture_file_path('onehundredrows.csv')
    project = Project('p-id')
    new_dataset = project.upload_dataset(filepath)
    assert_matches_expected_dataset(new_dataset)
    assert 'file' in responses.calls[0].request.body.fields


@responses.activate
@pytest.mark.usefixtures('client')
def test_delete_dataset():
    responses.add(responses.DELETE,
                  'https://host_name.com/projects/p-id/predictionDatasets/ds-id/',
                  body='')
    dataset = PredictionDataset.from_server_data(dataset_response)
    dataset.delete()
    assert responses.calls[0].request.method == 'DELETE'
    assert len(responses.calls) == 1



"""About datarobot
=========================
.. image:: https://img.shields.io/pypi/v/datarobot.svg
   :target: https://pypi.python.org/pypi/datarobot/
.. image:: https://img.shields.io/pypi/pyversions/datarobot.svg
.. image:: https://img.shields.io/pypi/status/datarobot.svg

DataRobot is a client library for working with the `DataRobot`_ platform API.

Installation
=========================
Python 2.7 and >= 3.4 are supported.
You must have a datarobot account.

::

   $ pip install datarobot

Usage
=========================
The library will look for a config file `~/.config/datarobot/drconfig.yaml` by default.
This is an example of what that config file should look like.

::

   token: your_token
   endpoint: https://app.datarobot.com/api/v2

Alternatively a global client can be set in the code.

::

   import datarobot as dr
   dr.Client(token='your_token', endpoint='https://app.datarobot.com/api/v2')

Alternatively environment variables can be used.

::

   export DATAROBOT_API_TOKEN='your_token'
   export DATAROBOT_ENDPOINT='https://app.datarobot.com/api/v2'

See `documentation`_ for example usage after configuring.

Tests
=========================
::

   $ py.test

.. _datarobot: http://datarobot.com
.. _documentation: https://datarobot-public-api-client.readthedocs-hosted.com
"""

import re
from setuptools import setup

with open('datarobot/_version.py') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

tests_require = ['mock', 'pytest-cov', 'responses']

dev_require = tests_require + [
    'flake8',
    'green',
    'ipython>=4.1,<5',
    'sphinx',
    'nbsphinx>=0.2.9,<1',
    'nbconvert>=5.3,<6',
    'numpydoc>=0.6.0',
    'tox',
]

example_require = [
    'ipython<=5.0',  # python2.7 compatablilty
    'jupyter<=5.0',
    'fredapi==0.4.0',
    'matplotlib<=2.0.2',
    'seaborn<=0.8',
    'sklearn<=0.18.2',
    'wordcloud<=1.3.1',
    'colour<=0.1.4'
]

setup(
    name='datarobot',
    version=version,
    description='This client library is designed to support the Datarobot API.',
    author='datarobot',
    author_email='support@datarobot.com',
    maintainer='datarobot',
    maintainer_email='info@datarobot.com',
    url='http://datarobot.com',
    license='Apache Software License',
    packages=[
        'datarobot',
        'datarobot.models',
        'datarobot.utils',
        'datarobot.helpers',
        'datarobot.ext',
    ],
    long_description=__doc__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=[
        'pandas>=0.15',
        'pyyaml>=3.11',
        'requests_toolbelt>=0.6',
        'trafaret>=0.7,<1.1',
    ],
    setup_requires=['pytest-runner'],
    tests_require=tests_require,
    extras_require={
        'dev': dev_require,
        'examples': example_require
    }
)



from .partitioning_methods import *  # noqa


class AdvancedOptions(object):
    """
    Used when setting the target of a project to set advanced options of modeling process.

    Parameters
    ----------
    weights : string, optional
        The name of a column indicating the weight of each row
    response_cap : float in [0.5, 1), optional
        Quantile of the response distribution to use for response capping.
    blueprint_threshold : int, optional
        Number of hours models are permitted to run before being excluded from later autopilot
        stages
        Minimum 1
    seed : int
        a seed to use for randomization
    smart_downsampled : bool
        whether to use smart downsampling to throw away excess rows of the majority class.  Only
        applicable to classification and zero-boosted regression projects.
    majority_downsampling_rate : float
        the percentage between 0 and 100 of the majority rows that should be kept.  Specify only if
        using smart downsampling.  May not cause the majority class to become smaller than the
        minority class.
    offset : list of str, optional
        (New in version v2.6) the list of the names of the columns containing the offset
        of each row
    exposure : string, optional
        (New in version v2.6) the name of a column containing the exposure of each row
    accuracy_optimized_mb : bool, optional
        (New in version v2.6) Include additional, longer-running models that will be run by the
        autopilot and available to run manually.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        advanced_options = dr.AdvancedOptions(
            weights='weights_column',
            offset=['offset_column'],
            exposure='exposure_column',
            response_cap=0.7,
            blueprint_threshold=2,
            smart_downsampled=True, majority_downsampling_rate=75.0)

    """
    def __init__(self, weights=None, response_cap=None, blueprint_threshold=None,
                 seed=None, smart_downsampled=False, majority_downsampling_rate=None,
                 offset=None, exposure=None, accuracy_optimized_mb=None):
        self.weights = weights
        self.response_cap = response_cap
        self.blueprint_threshold = blueprint_threshold
        self.seed = seed
        self.smart_downsampled = smart_downsampled
        self.majority_downsampling_rate = majority_downsampling_rate
        self.offset = offset
        self.exposure = exposure
        self.accuracy_optimized_mb = accuracy_optimized_mb

    def collect_payload(self):

        payload = dict(
            weights=self.weights,
            response_cap=self.response_cap,
            blueprint_threshold=self.blueprint_threshold,
            seed=self.seed,
            smart_downsampled=self.smart_downsampled,
            majority_downsampling_rate=self.majority_downsampling_rate,
            offset=self.offset,
            exposure=self.exposure,
            accuracy_optimized_mb=self.accuracy_optimized_mb
        )
        return payload
